import copy
import math
from typing import List, Optional, Tuple, Union

import torch
from torch import Tensor, nn
from torch.nn import CrossEntropyLoss
from transformers import AutoModel, AutoModelForTokenClassification
from transformers.modeling_outputs import (
    BaseModelOutputWithPoolingAndCrossAttentions,
    BaseModelOutputWithPastAndCrossAttentions,
)
from transformers.modeling_utils import ModuleUtilsMixin
from transformers.models.bert.modeling_bert import BertEncoder, BertForTokenClassification, BertModel, BertPooler
from transformers.models.canine.modeling_canine import (
    _PRIMES,
    ACT2FN,
    BaseModelOutput,
    CanineAttention,
    CanineEmbeddings,
    CanineEncoder,
    CanineForTokenClassification,
    CanineIntermediate,
    CanineLayer,
    CanineModel,
    CanineModelOutputWithPooling,
    CanineOutput,
    CaninePooler,
    CanineSelfAttention,
    CanineSelfOutput,
    CharactersToMolecules,
    ConvProjection,
    TokenClassifierOutput,
)
from transformers.models.xlm_roberta import (
    XLMRobertaForTokenClassification,
    XLMRobertaModel,
)
from transformers.models.xlm_roberta.modeling_xlm_roberta import (
    XLMRobertaEmbeddings,
    XLMRobertaPooler,
    XLMRobertaLayer,
)

from wtpsplit.configs import BertCharConfig, LACanineConfig, SubwordXLMConfig
from wtpsplit.utils import Constants


# added n-gram representations
class LACanineEmbeddings(CanineEmbeddings):
    def __init__(self, config):
        super().__init__(config)

        self.ngram_order = getattr(config, "ngram_order", 1)
        if self.ngram_order > 1:
            raise NotImplementedError("n-gram representations are not implemented.")

        shard_embedding_size = config.hidden_size // config.num_hash_functions
        for j in range(2, self.ngram_order + 1):
            for i in range(config.num_hash_functions):
                name = f"HashBucketCodepointEmbedder_{i}_{j}"
                setattr(
                    self,
                    name,
                    nn.Embedding(config.num_hash_buckets, shard_embedding_size),
                )

    def _embed_hash_buckets(self, input_ids=None, hashed_ids=None):
        embedding_size = self.config.hidden_size
        num_hashes = self.config.num_hash_functions
        num_buckets = self.config.num_hash_buckets

        assert (input_ids is None) + (
            hashed_ids is None
        ) == 1, "Either `input_ids` or `hashed_ids` must be provided (and not both!)."

        """Converts IDs (e.g. codepoints) into embeddings via multiple hashing."""
        if embedding_size % num_hashes != 0:
            raise ValueError(f"Expected `embedding_size` ({embedding_size}) % `num_hashes` ({num_hashes}) == 0")

        if num_hashes > len(_PRIMES):
            raise ValueError(f"`num_hashes` must be <= {len(_PRIMES)}")

        embedding_shards = []
        for i in range(num_hashes):
            name = f"HashBucketCodepointEmbedder_{i}"

            hash_ids = ((input_ids + 1) * _PRIMES[i]) % num_buckets if hashed_ids is None else hashed_ids[:, :, i]

            shard_embeddings = getattr(self, name)(hash_ids)
            embedding_shards.append(shard_embeddings)

        return torch.cat(embedding_shards, dim=-1)

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        hashed_ids: Optional[torch.LongTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        **kwargs,
    ) -> torch.FloatTensor:
        if input_ids is not None:
            input_shape = input_ids.size()
        elif hashed_ids is not None:
            input_shape = hashed_ids.size()[:-1]
        else:
            input_shape = inputs_embeds.size()[:-1]

        seq_length = input_shape[1]

        if position_ids is None:
            position_ids = self.position_ids[:, :seq_length]

        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=self.position_ids.device)

        if inputs_embeds is None:
            inputs_embeds = self._embed_hash_buckets(input_ids, hashed_ids)

        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = inputs_embeds + token_type_embeddings

        if self.position_embedding_type == "absolute":
            position_embeddings = self.char_position_embeddings(position_ids)
            embeddings += position_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class LACanineSelfAttention(CanineSelfAttention):
    def __init__(self, config):
        super(CanineSelfAttention, self).__init__()
        if config.hidden_size % config.num_attention_heads != 0 and not hasattr(config, "embedding_size"):
            raise ValueError(
                f"The hidden size ({config.hidden_size}) is not a multiple of the number of attention "
                f"heads ({config.num_attention_heads})"
            )

        self.lookahead = config.lookahead
        self.lookahead_block_size = config.lookahead_block_size

        if self.lookahead is not None:
            max_positions = config.max_position_embeddings
            assert max_positions % self.lookahead_block_size == 0

            self.register_buffer(
                "bias",
                torch.tril(
                    torch.ones(
                        (
                            max_positions // self.lookahead_block_size,
                            max_positions // self.lookahead_block_size,
                        ),
                        dtype=torch.uint8,
                    ),
                    diagonal=self.lookahead,
                )
                .repeat_interleave(self.lookahead_block_size, dim=1)
                .repeat_interleave(self.lookahead_block_size, dim=0)
                .view(
                    1,
                    1,
                    max_positions,
                    max_positions,
                ),
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.position_embedding_type = getattr(config, "position_embedding_type", "absolute")
        if self.position_embedding_type == "relative_key" or self.position_embedding_type == "relative_key_query":
            self.max_position_embeddings = config.max_position_embeddings
            self.distance_embedding = nn.Embedding(2 * config.max_position_embeddings - 1, self.attention_head_size)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (
            self.num_attention_heads,
            self.attention_head_size,
        )
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
        self,
        from_tensor: torch.Tensor,
        to_tensor: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        mixed_query_layer = self.query(from_tensor)

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.

        key_layer = self.transpose_for_scores(self.key(to_tensor))
        value_layer = self.transpose_for_scores(self.value(to_tensor))

        query_layer = self.transpose_for_scores(mixed_query_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

        if self.position_embedding_type == "relative_key" or self.position_embedding_type == "relative_key_query":
            seq_length = from_tensor.size()[1]
            position_ids_l = torch.arange(seq_length, dtype=torch.long, device=from_tensor.device).view(-1, 1)
            position_ids_r = torch.arange(seq_length, dtype=torch.long, device=from_tensor.device).view(1, -1)
            distance = position_ids_l - position_ids_r
            positional_embedding = self.distance_embedding(distance + self.max_position_embeddings - 1)
            positional_embedding = positional_embedding.to(dtype=query_layer.dtype)  # fp16 compatibility

            if self.position_embedding_type == "relative_key":
                relative_position_scores = torch.einsum("bhld,lrd->bhlr", query_layer, positional_embedding)
                attention_scores = attention_scores + relative_position_scores
            elif self.position_embedding_type == "relative_key_query":
                relative_position_scores_query = torch.einsum("bhld,lrd->bhlr", query_layer, positional_embedding)
                relative_position_scores_key = torch.einsum("bhrd,lrd->bhlr", key_layer, positional_embedding)
                attention_scores = attention_scores + relative_position_scores_query + relative_position_scores_key

        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            if attention_mask.ndim == 3:
                # if attention_mask is 3D, do the following:
                attention_mask = torch.unsqueeze(attention_mask, dim=1)
                # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
                # masked positions, this operation will create a tensor which is 0.0 for
                # positions we want to attend and -10000.0 for masked positions.
                attention_mask = (1.0 - attention_mask.to(attention_scores.dtype)) * torch.finfo(
                    attention_scores.dtype
                ).min

            # Apply the attention mask (precomputed for all layers in CanineModel forward() function)
            attention_scores = attention_scores + attention_mask

        if self.lookahead is not None:
            query_length, key_length = query_layer.size(-2), key_layer.size(-2)
            causal_mask = self.bias[
                :,
                :,
                key_length - query_length : key_length,
                :key_length,
            ].to(torch.bool)
            mask_value = torch.finfo(attention_scores.dtype).min
            mask_value = torch.tensor(mask_value, dtype=attention_scores.dtype).to(attention_scores.device)
            attention_scores = torch.where(
                causal_mask,
                attention_scores,
                mask_value,
            )

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

        return outputs


class LACanineAttention(CanineAttention):
    """
    Additional arguments related to local attention:

        - **local** (`bool`, *optional*, defaults to `False`) -- Whether to apply local attention.
        - **always_attend_to_first_position** (`bool`, *optional*, defaults to `False`) -- Should all blocks be able to
          attend
        to the `to_tensor`'s first position (e.g. a [CLS] position)? - **first_position_attends_to_all** (`bool`,
        *optional*, defaults to `False`) -- Should the *from_tensor*'s first position be able to attend to all
        positions within the *from_tensor*? - **attend_from_chunk_width** (`int`, *optional*, defaults to 128) -- The
        width of each block-wise chunk in `from_tensor`. - **attend_from_chunk_stride** (`int`, *optional*, defaults to
        128) -- The number of elements to skip when moving to the next block in `from_tensor`. -
        **attend_to_chunk_width** (`int`, *optional*, defaults to 128) -- The width of each block-wise chunk in
        *to_tensor*. - **attend_to_chunk_stride** (`int`, *optional*, defaults to 128) -- The number of elements to
        skip when moving to the next block in `to_tensor`.
    """

    def __init__(
        self,
        config,
        local=False,
        always_attend_to_first_position: bool = False,
        first_position_attends_to_all: bool = False,
        attend_from_chunk_width: int = 128,
        attend_from_chunk_stride: int = 128,
        attend_to_chunk_width: int = 128,
        attend_to_chunk_stride: int = 128,
    ):
        super(CanineAttention, self).__init__()
        self.self = LACanineSelfAttention(config)
        self.output = CanineSelfOutput(config)
        self.pruned_heads = set()

        # additional arguments related to local attention
        self.local = local
        if attend_from_chunk_width < attend_from_chunk_stride:
            raise ValueError(
                "`attend_from_chunk_width` < `attend_from_chunk_stride` would cause sequence positions to get skipped."
            )
        if attend_to_chunk_width < attend_to_chunk_stride:
            raise ValueError(
                "`attend_to_chunk_width` < `attend_to_chunk_stride`would cause sequence positions to get skipped."
            )
        self.always_attend_to_first_position = always_attend_to_first_position
        self.first_position_attends_to_all = first_position_attends_to_all
        self.attend_from_chunk_width = attend_from_chunk_width
        self.attend_from_chunk_stride = attend_from_chunk_stride
        self.attend_to_chunk_width = attend_to_chunk_width
        self.attend_to_chunk_stride = attend_to_chunk_stride


class LACanineOutput(CanineOutput):
    def __init__(self, config):
        super().__init__(config)
        self.bottleneck_size = config.hidden_size // config.bottleneck_factor
        self.hidden_size = config.hidden_size

        if isinstance(config.hidden_act, str):
            self.act_fn = ACT2FN[config.hidden_act]
        else:
            self.act_fn = config.hidden_act

        self.language_adapter = config.language_adapter

        if self.language_adapter in {"on", "shared"}:
            # language adapter init
            dummy_ff_downs = [nn.Linear(self.hidden_size, self.bottleneck_size) for _ in range(config.n_languages)]
            dummy_ff_ups = [nn.Linear(self.bottleneck_size, self.hidden_size) for _ in range(config.n_languages)]

            self.lang_ff_down_weights = nn.Embedding(
                config.n_languages,
                (self.hidden_size * self.bottleneck_size),
                _weight=torch.stack([l.weight.data.T.reshape(-1) for l in dummy_ff_downs]),
            )
            self.lang_ff_up_weights = nn.Embedding(
                config.n_languages,
                (self.bottleneck_size * self.hidden_size),
                _weight=torch.stack([l.weight.data.T.reshape(-1) for l in dummy_ff_ups]),
            )
            self.lang_ff_down_biases = nn.Embedding(
                config.n_languages,
                self.bottleneck_size,
                _weight=torch.stack([l.bias.data for l in dummy_ff_downs]),
            )
            self.lang_ff_up_biases = nn.Embedding(
                config.n_languages,
                self.hidden_size,
                _weight=torch.stack([l.bias.data for l in dummy_ff_ups]),
            )

    def forward(
        self,
        hidden_states: Tuple[torch.FloatTensor],
        input_tensor: torch.FloatTensor,
        language_ids: Optional[torch.LongTensor],
    ) -> torch.FloatTensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)

        hidden_states = self.LayerNorm(hidden_states + input_tensor)

        if self.language_adapter == "off":
            return hidden_states
        elif self.language_adapter == "on":
            ff_down_weights = self.lang_ff_down_weights(language_ids)
            ff_up_weights = self.lang_ff_up_weights(language_ids)
            ff_down_biases = self.lang_ff_down_biases(language_ids)
            ff_up_biases = self.lang_ff_up_biases(language_ids)
        elif self.language_adapter == "shared":
            ff_down_weights = self.lang_ff_down_weights.weight.mean(0, keepdim=True)
            ff_up_weights = self.lang_ff_up_weights.weight.mean(0, keepdim=True)
            ff_down_biases = self.lang_ff_down_biases.weight.mean(0, keepdim=True)
            ff_up_biases = self.lang_ff_up_biases.weight.mean(0, keepdim=True)

        ff_down_weights = ff_down_weights.view(-1, self.hidden_size, self.bottleneck_size)
        ff_up_weights = ff_down_weights.view(-1, self.bottleneck_size, self.hidden_size)

        down = self.act_fn(torch.matmul(hidden_states, ff_down_weights) + ff_down_biases.unsqueeze(1))
        up = torch.matmul(down, ff_up_weights) + ff_up_biases.unsqueeze(1)

        hidden_states = self.LayerNorm(hidden_states + up)

        return hidden_states


class LACanineLayer(CanineLayer):
    def __init__(
        self,
        config,
        local,
        always_attend_to_first_position,
        first_position_attends_to_all,
        attend_from_chunk_width,
        attend_from_chunk_stride,
        attend_to_chunk_width,
        attend_to_chunk_stride,
    ):
        super(CanineLayer, self).__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        assert self.chunk_size_feed_forward == 0
        self.seq_len_dim = 1
        self.attention = LACanineAttention(
            config,
            local,
            always_attend_to_first_position,
            first_position_attends_to_all,
            attend_from_chunk_width,
            attend_from_chunk_stride,
            attend_to_chunk_width,
            attend_to_chunk_stride,
        )
        self.intermediate = CanineIntermediate(config)
        self.output = LACanineOutput(config)

    def forward(
        self,
        hidden_states: Tuple[torch.FloatTensor],
        language_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.FloatTensor, Optional[torch.FloatTensor]]:
        self_attention_outputs = self.attention(
            hidden_states,
            attention_mask,
            head_mask,
            output_attentions=output_attentions,
        )
        attention_output = self_attention_outputs[0]

        outputs = self_attention_outputs[1:]  # add self attentions if we output attention weights

        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output, language_ids)

        outputs = (layer_output,) + outputs

        return outputs

    def feed_forward_chunk(self, attention_output):
        raise NotImplementedError()


class LACanineEncoder(CanineEncoder):
    def __init__(
        self,
        config,
        local=False,
        always_attend_to_first_position=False,
        first_position_attends_to_all=False,
        attend_from_chunk_width=128,
        attend_from_chunk_stride=128,
        attend_to_chunk_width=128,
        attend_to_chunk_stride=128,
    ):
        super(CanineEncoder, self).__init__()
        self.config = config
        self.layer = nn.ModuleList(
            [
                LACanineLayer(
                    config,
                    local,
                    always_attend_to_first_position,
                    first_position_attends_to_all,
                    attend_from_chunk_width,
                    attend_from_chunk_stride,
                    attend_to_chunk_width,
                    attend_to_chunk_stride,
                )
                for _ in range(config.num_hidden_layers)
            ]
        )
        self.gradient_checkpointing = False

    def forward(
        self,
        hidden_states: Tuple[torch.FloatTensor],
        language_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
        output_hidden_states: Optional[bool] = False,
        return_dict: Optional[bool] = True,
    ) -> Union[Tuple, BaseModelOutput]:
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None

        for i, layer_module in enumerate(self.layer):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None

            if self.gradient_checkpointing and self.training:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs, output_attentions)

                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(layer_module),
                    hidden_states,
                    language_ids,
                    attention_mask,
                    layer_head_mask,
                )
            else:
                layer_outputs = layer_module(
                    hidden_states,
                    language_ids,
                    attention_mask,
                    layer_head_mask,
                    output_attentions,
                )

            hidden_states = layer_outputs[0]
            if output_attentions:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, all_hidden_states, all_self_attentions] if v is not None)
        return BaseModelOutput(
            last_hidden_state=hidden_states,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


class CausalCharactersToMolecules(CharactersToMolecules):
    """Convert character sequence to initial molecule sequence (i.e. downsample) using strided convolutions."""

    def forward(self, char_encoding: torch.Tensor) -> torch.Tensor:
        # MODIFIED: no cls encoding!
        assert char_encoding.shape[1] % self.conv.stride[0] == 0

        # char_encoding has shape [batch, char_seq, hidden_size]
        # We transpose it to be [batch, hidden_size, char_seq]
        char_encoding = torch.transpose(char_encoding, 1, 2)
        downsampled = self.conv(char_encoding)
        downsampled = torch.transpose(downsampled, 1, 2)
        downsampled = self.activation(downsampled)

        result = downsampled

        result = self.LayerNorm(result)

        return result


class LACanineModel(CanineModel):
    config_class = LACanineConfig

    def __init__(self, config, add_pooling_layer=True):
        super(CanineModel, self).__init__(config)
        self.config = config
        final_shallow_config = copy.deepcopy(config)
        final_shallow_config.num_hidden_layers = 1

        initial_shallow_config = copy.deepcopy(config)
        initial_shallow_config.num_hidden_layers = 1

        if config.lookahead is not None:
            initial_shallow_config.lookahead = config.lookahead
            initial_shallow_config.lookahead_block_size = config.downsampling_rate

            # the ConvProjection must not leak future context (ks = 1)
            final_shallow_config.lookahead_block_size = config.downsampling_rate
            final_shallow_config.lookahead = 0

            config.lookahead = 0

        self.char_embeddings = LACanineEmbeddings(config)
        # shallow/low-dim transformer encoder to get a initial character encoding
        self.initial_char_encoder = LACanineEncoder(
            initial_shallow_config,
            local=True,
            always_attend_to_first_position=False,
            first_position_attends_to_all=False,
            attend_from_chunk_width=config.local_transformer_stride,
            attend_from_chunk_stride=config.local_transformer_stride,
            attend_to_chunk_width=config.local_transformer_stride,
            attend_to_chunk_stride=config.local_transformer_stride,
        )
        self.chars_to_molecules = CausalCharactersToMolecules(config)
        # deep transformer encoder
        self.encoder = LACanineEncoder(config)
        self.projection = ConvProjection(config)
        # shallow/low-dim transformer encoder to get a final character encoding
        self.final_char_encoder = LACanineEncoder(final_shallow_config)

        self.pooler = CaninePooler(config) if add_pooling_layer else None

        # Initialize weights and apply final processing
        self.post_init()

    def _repeat_molecules(self, molecules: torch.Tensor, char_seq_length: torch.Tensor) -> torch.Tensor:
        """Repeats molecules to make them the same length as the char sequence."""

        rate = self.config.downsampling_rate

        return torch.repeat_interleave(molecules, repeats=rate, dim=-2)

    def _downsample_attention_mask(self, char_attention_mask: torch.Tensor, downsampling_rate: int):
        """Downsample 2D character attention mask to 2D molecule attention mask using MaxPool1d layer."""

        # if broadcasted, use MaxPool1d to just pool the last dimension
        if char_attention_mask.shape[2] == 1:
            return torch.nn.MaxPool1d(kernel_size=downsampling_rate, stride=downsampling_rate)(
                char_attention_mask.squeeze(2)
            ).unsqueeze(2)

        return torch.nn.MaxPool2d(kernel_size=downsampling_rate, stride=downsampling_rate)(char_attention_mask)

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        language_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CanineModelOutputWithPooling]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        batch_size, seq_length = input_shape
        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if attention_mask is None:
            attention_mask = torch.ones(((batch_size, seq_length)), device=device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape)
        extended_molecule_attention_mask = self._downsample_attention_mask(
            extended_attention_mask, downsampling_rate=self.config.downsampling_rate
        )

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        # `input_char_embeddings`: shape (batch_size, char_seq, char_dim)
        input_char_embeddings = self.char_embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
        )

        # Contextualize character embeddings using shallow Transformer.
        # We use a 3D attention mask for the local attention.
        # `input_char_encoding`: shape (batch_size, char_seq_len, char_dim)
        if attention_mask.ndim == 2:
            char_attention_mask = self._create_3d_attention_mask_from_input_mask(
                input_ids or inputs_embeds, attention_mask
            )
        else:
            char_attention_mask = attention_mask
        init_chars_encoder_outputs = self.initial_char_encoder(
            input_char_embeddings,
            language_ids=language_ids,
            attention_mask=char_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        input_char_encoding = init_chars_encoder_outputs.last_hidden_state

        # Downsample chars to molecules.
        # The following lines have dimensions: [batch, molecule_seq, molecule_dim].
        # In this transformation, we change the dimensionality from `char_dim` to
        # `molecule_dim`, but do *NOT* add a resnet connection. Instead, we rely on
        # the resnet connections (a) from the final char transformer stack back into
        # the original char transformer stack and (b) the resnet connections from
        # the final char transformer stack back into the deep BERT stack of
        # molecules.
        #
        # Empirically, it is critical to use a powerful enough transformation here:
        # mean pooling causes training to diverge with huge gradient norms in
        # this region of the model; using a convolution here resolves this issue. From
        # this, it seems that molecules and characters require a very different
        # feature space; intuitively, this makes sense.
        init_molecule_encoding = self.chars_to_molecules(input_char_encoding)

        # Deep BERT encoder
        # `molecule_sequence_output`: shape (batch_size, mol_seq_len, mol_dim)
        encoder_outputs = self.encoder(
            init_molecule_encoding,
            language_ids=language_ids,
            attention_mask=extended_molecule_attention_mask,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        molecule_sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(molecule_sequence_output) if self.pooler is not None else None

        # Upsample molecules back to characters.
        # `repeated_molecules`: shape (batch_size, char_seq_len, mol_hidden_size)
        repeated_molecules = self._repeat_molecules(molecule_sequence_output, char_seq_length=input_shape[-1])

        # Concatenate representations (contextualized char embeddings and repeated molecules):
        # `concat`: shape [batch_size, char_seq_len, molecule_hidden_size+char_hidden_final]
        concat = torch.cat([input_char_encoding, repeated_molecules], dim=-1)

        # Project representation dimension back to hidden_size
        # `sequence_output`: shape (batch_size, char_seq_len, hidden_size])
        sequence_output = self.projection(concat)

        # Apply final shallow Transformer
        # `sequence_output`: shape (batch_size, char_seq_len, hidden_size])
        final_chars_encoder_outputs = self.final_char_encoder(
            sequence_output,
            language_ids=language_ids,
            attention_mask=extended_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        sequence_output = final_chars_encoder_outputs.last_hidden_state

        if output_hidden_states:
            deep_encoder_hidden_states = encoder_outputs.hidden_states if return_dict else encoder_outputs[1]
            all_hidden_states = (
                all_hidden_states
                + init_chars_encoder_outputs.hidden_states
                + deep_encoder_hidden_states
                + final_chars_encoder_outputs.hidden_states
            )

        if output_attentions:
            deep_encoder_self_attentions = encoder_outputs.attentions if return_dict else encoder_outputs[-1]
            all_self_attentions = (
                all_self_attentions
                + init_chars_encoder_outputs.attentions
                + deep_encoder_self_attentions
                + final_chars_encoder_outputs.attentions
            )

        if not return_dict:
            output = (sequence_output, pooled_output)
            output += tuple(v for v in [all_hidden_states, all_self_attentions] if v is not None)
            return output

        return CanineModelOutputWithPooling(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


class LACanineForTokenClassification(CanineForTokenClassification):
    config_class = LACanineConfig

    def __init__(self, config):
        super(CanineForTokenClassification, self).__init__(config)
        self.num_labels = config.num_labels

        self.canine = LACanineModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    def set_language_adapters(self, lang):
        lang_index = Constants.LANG_CODE_TO_INDEX[lang]

        for module in self.modules():
            if isinstance(module, LACanineOutput):
                module.lang_ff_down_biases.weight = nn.Parameter(module.lang_ff_down_biases.weight[[lang_index]])
                module.lang_ff_down_weights.weight = nn.Parameter(module.lang_ff_down_weights.weight[[lang_index]])
                module.lang_ff_up_biases.weight = nn.Parameter(module.lang_ff_up_biases.weight[[lang_index]])
                module.lang_ff_up_weights.weight = nn.Parameter(module.lang_ff_up_weights.weight[[lang_index]])
                module.language_adapter = "shared"

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        language_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        hashed_ids: Optional[torch.Tensor] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, TokenClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for computing the token classification loss. Indices should be in `[0, ..., config.num_labels - 1]`.
        """
        inputs_embeds = self.canine.char_embeddings._embed_hash_buckets(
            input_ids=input_ids,
            hashed_ids=hashed_ids,
        )
        input_ids = None

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.canine(
            input_ids,
            language_ids=language_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        sequence_output = outputs[0]

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        loss = None
        if labels is not None:
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class BertCharModel(BertModel):
    config_class = BertCharConfig

    """
    The model can behave as an encoder (with only self-attention) as well as a decoder, in which case a layer of
    cross-attention is added between the self-attention layers, following the architecture described in [Attention is
    all you need](https://arxiv.org/abs/1706.03762) by Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit,
    Llion Jones, Aidan N. Gomez, Lukasz Kaiser and Illia Polosukhin.
    To behave as an decoder the model needs to be initialized with the `is_decoder` argument of the configuration set
    to `True`. To be used in a Seq2Seq model, the model needs to initialized with both `is_decoder` argument and
    `add_cross_attention` set to `True`; an `encoder_hidden_states` is then expected as an input to the forward pass.
    """

    def __init__(self, config, add_pooling_layer=True):
        super().__init__(config)
        self.config = config

        self.embeddings = LACanineEmbeddings(config)
        self.encoder = BertEncoder(config)

        self.pooler = BertPooler(config) if add_pooling_layer else None

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.embeddings


class BertCharForTokenClassification(BertForTokenClassification):
    config_class = BertCharConfig

    _keys_to_ignore_on_load_unexpected = [r"pooler"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertCharModel(config, add_pooling_layer=False)
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        hashed_ids: Optional[torch.Tensor] = None,
        language_ids=None,
        return_dict: Optional[bool] = None,
    ):
        inputs_embeds = self.bert.embeddings._embed_hash_buckets(
            input_ids=input_ids,
            hashed_ids=hashed_ids,
        )
        input_ids = None

        return super().forward(
            input_ids,
            attention_mask,
            token_type_ids,
            position_ids,
            head_mask,
            inputs_embeds,
            labels,
            output_attentions,
            output_hidden_states,
            return_dict,
        )


class SubwordXLMForTokenClassification(XLMRobertaForTokenClassification):
    config_class = SubwordXLMConfig

    _keys_to_ignore_on_load_unexpected = [r"pooler"]
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.roberta = SubwordXLMRobertaModel(config, add_pooling_layer=False)
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        hashed_ids: Optional[torch.Tensor] = None,
        language_ids=None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], TokenClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for computing the token classification loss. Indices should be in `[0, ..., config.num_labels - 1]`.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        sequence_output = outputs[0]

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        loss = None
        if labels is not None:
            # move labels to correct device to enable model parallelism
            labels = labels.to(logits.device)
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class SubwordXLMRobertaModel(XLMRobertaModel):
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    # Copied from transformers.models.bert.modeling_bert.BertModel.__init__ with Bert->XLMRoberta
    def __init__(self, config, add_pooling_layer=True):
        super().__init__(config)
        self.config = config

        self.embeddings = XLMRobertaEmbeddings(config)
        self.encoder = SubwordXLMRobertaEncoder(config)
        self.lookahead_split_layers = config.lookahead_split_layers
        self.pooler = XLMRobertaPooler(config) if add_pooling_layer else None
        if config.lookahead is not None:
            divide_value = (
                self.lookahead_split_layers if self.lookahead_split_layers is not None else config.num_hidden_layers
            )
            if config.lookahead_split_layers is not None:
                assert self.lookahead_split_layers <= config.num_hidden_layers
            self.effective_lookahead = config.lookahead // divide_value
        else:
            self.effective_lookahead = None

        # Initialize weights and apply final processing
        self.post_init()

    # Copied from transformers.models.bert.modeling_bert.BertModel.forward
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], BaseModelOutputWithPoolingAndCrossAttentions]:
        r"""
        encoder_hidden_states  (`torch.FloatTensor` of shape `(batch_size, sequence_length, hidden_size)`, *optional*):
            Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention if
            the model is configured as a decoder.
        encoder_attention_mask (`torch.FloatTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Mask to avoid performing attention on the padding token indices of the encoder input. This mask is used in
            the cross-attention if the model is configured as a decoder. Mask values selected in `[0, 1]`:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        past_key_values (`tuple(tuple(torch.FloatTensor))` of length `config.n_layers` with each tuple having 4 tensors of shape `(batch_size, num_heads, sequence_length - 1, embed_size_per_head)`):
            Contains precomputed key and value hidden states of the attention blocks. Can be used to speed up decoding.

            If `past_key_values` are used, the user can optionally input only the last `decoder_input_ids` (those that
            don't have their past key value states given to this model) of shape `(batch_size, 1)` instead of all
            `decoder_input_ids` of shape `(batch_size, sequence_length)`.
        use_cache (`bool`, *optional*):
            If set to `True`, `past_key_values` key value states are returned and can be used to speed up decoding (see
            `past_key_values`).
        """
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if self.config.is_decoder:
            use_cache = use_cache if use_cache is not None else self.config.use_cache
        else:
            use_cache = False

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        batch_size, seq_length = input_shape
        device = input_ids.device if input_ids is not None else inputs_embeds.device

        # past_key_values_length
        past_key_values_length = past_key_values[0][0].shape[2] if past_key_values is not None else 0

        if attention_mask is None:
            attention_mask = torch.ones(((batch_size, seq_length + past_key_values_length)), device=device)

        if token_type_ids is None:
            if hasattr(self.embeddings, "token_type_ids"):
                buffered_token_type_ids = self.embeddings.token_type_ids[:, :seq_length]
                buffered_token_type_ids_expanded = buffered_token_type_ids.expand(batch_size, seq_length)
                token_type_ids = buffered_token_type_ids_expanded
            else:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = get_extended_attention_mask(
            self.config, attention_mask, input_shape, self.effective_lookahead, device, self.dtype
        )

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if self.config.is_decoder and encoder_hidden_states is not None:
            encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
            if encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
            encoder_extended_attention_mask = self.invert_attention_mask(encoder_attention_mask)
        else:
            encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
        )
        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            init_attention_mask=attention_mask,
            dtype=self.dtype,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


def get_extended_attention_mask(
    config,
    attention_mask: Tensor,
    input_shape: Tuple[int],
    lookahead: Optional[int] = None,
    device: torch.device = None,
    dtype: torch.float = None,
) -> Tensor:
    """
    Makes broadcastable attention and causal masks so that future and masked tokens are ignored.

    Arguments:
        attention_mask (`torch.Tensor`):
            Mask with ones indicating tokens to attend to, zeros for tokens to ignore.
        input_shape (`Tuple[int]`):
            The shape of the input to the model.

    Returns:
        `torch.Tensor` The extended attention mask, with a the same dtype as `attention_mask.dtype`.
    """

    # if not (attention_mask.dim() == 2 and config.is_decoder):
    # show warning only if it won't be shown in `create_extended_attention_mask_for_decoder`
    # if device is not None:
    #     warnings.warn(
    #         "The `device` argument is deprecated and will be removed in v5 of Transformers.", FutureWarning
    #     )
    # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
    # ourselves in which case we just need to make it broadcastable to all heads.
    if attention_mask.dim() == 3:
        extended_attention_mask = attention_mask[:, None, :, :]

    elif attention_mask.dim() == 2:
        # Provided a padding mask of dimensions [batch_size, seq_length]
        # - if the model is a decoder, apply a causal mask in addition to the padding mask
        # - if the model is an encoder, make the mask broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if config.is_decoder:
            extended_attention_mask = ModuleUtilsMixin.create_extended_attention_mask_for_decoder(
                input_shape, attention_mask, device
            )
        if lookahead is not None:
            # lookahead mask of shape [batch_size, 1, seq_length, seq_length]
            # the current token should attend to the next `lookahead` tokens
            # the current token should not attend to the previous `lookahead` tokens
            _, seq_length = attention_mask.shape
            # Create a lookahead mask
            lookahead_mask = torch.tril(torch.ones(seq_length, seq_length), diagonal=lookahead, out=None).to(
                attention_mask.device
            )
            # Combine the attention mask with the lookahead mask
            extended_attention_mask = attention_mask[:, None, None, :] * lookahead_mask
        else:
            extended_attention_mask = attention_mask[:, None, None, :]
    else:
        raise ValueError(
            f"Wrong shape for input_ids (shape {input_shape}) or attention_mask (shape {attention_mask.shape})"
        )

    # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
    # masked positions, this operation will create a tensor which is 0.0 for
    # positions we want to attend and the dtype's smallest value for masked positions.
    # Since we are adding it to the raw scores before the softmax, this is
    # effectively the same as removing these entirely.
    extended_attention_mask = extended_attention_mask.to(dtype=dtype)  # fp16 compatibility
    extended_attention_mask = (1.0 - extended_attention_mask) * torch.finfo(dtype).min
    return extended_attention_mask


# Copied from transformers.models.roberta.modeling_roberta.RobertaEncoder with Roberta->XLMRoberta
class SubwordXLMRobertaEncoder(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList([XLMRobertaLayer(config) for _ in range(config.num_hidden_layers)])
        self.gradient_checkpointing = False

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = False,
        output_hidden_states: Optional[bool] = False,
        return_dict: Optional[bool] = True,
        init_attention_mask: Optional[torch.Tensor] = None,
        dtype: torch.float = None,
    ) -> Union[Tuple[torch.Tensor], BaseModelOutputWithPastAndCrossAttentions]:
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None
        all_cross_attentions = () if output_attentions and self.config.add_cross_attention else None

        if self.gradient_checkpointing and self.training:
            if use_cache:
                print("`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`...")
                use_cache = False

        next_decoder_cache = () if use_cache else None
        for i, layer_module in enumerate(self.layer):
            # MODIFIED: if lookahead_split_layers is given, use causal mask starting from that layer
            if self.config.lookahead_split_layers is not None:
                if i == self.config.lookahead_split_layers:
                    attention_mask = get_extended_attention_mask(
                        self.config, init_attention_mask, init_attention_mask.shape, 0, hidden_states.device, dtype
                    )

            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None
            past_key_value = past_key_values[i] if past_key_values is not None else None

            if self.gradient_checkpointing and self.training:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs, past_key_value, output_attentions)

                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(layer_module),
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                )
            else:
                layer_outputs = layer_module(
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    past_key_value,
                    output_attentions,
                )

            hidden_states = layer_outputs[0]
            if use_cache:
                next_decoder_cache += (layer_outputs[-1],)
            if output_attentions:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)
                if self.config.add_cross_attention:
                    all_cross_attentions = all_cross_attentions + (layer_outputs[2],)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(
                v
                for v in [
                    hidden_states,
                    next_decoder_cache,
                    all_hidden_states,
                    all_self_attentions,
                    all_cross_attentions,
                ]
                if v is not None
            )
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=next_decoder_cache,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
            cross_attentions=all_cross_attentions,
        )


AutoModel.register(LACanineConfig, LACanineModel)
AutoModelForTokenClassification.register(LACanineConfig, LACanineForTokenClassification)

AutoModel.register(BertCharConfig, BertCharModel)
AutoModelForTokenClassification.register(BertCharConfig, BertCharForTokenClassification)

AutoModel.register(SubwordXLMConfig, SubwordXLMForTokenClassification)
AutoModelForTokenClassification.register(SubwordXLMConfig, SubwordXLMForTokenClassification)

# if __name__ == "__main__":
#     # test XLM
#     from transformers import AutoTokenizer

#     model_str = "xlm-roberta-base"
#     config = SubwordXLMConfig.from_pretrained(model_str)
#     config.num_labels = 4
#     config.num_hidden_layers = 12
#     config.lookahead = 48
#     config.lookahead_split_layers = 6
#     backbone = SubwordXLMForTokenClassification.from_pretrained(model_str, config=config)
#     print(summary(backbone, depth=4))

#     # some sample input
#     text = "A sentence. Now we move on. And on and this is the last sentence. Now, we are starting to move on to the next sentence. This is the last sentence."
#     tokenizer = AutoTokenizer.from_pretrained(model_str)

#     tokens = tokenizer(text, return_tensors="pt", add_special_tokens=False, pad_to_multiple_of=512, padding=True)
#     from tokenizers import AddedToken

#     tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})
#     print(tokenizer.tokenize(text))
#     print(tokenizer.encode(text))
#     print(tokens)

#     # forward pass
#     print(backbone(**tokens))



import glob
import gzip
import json
import os
import random
from dataclasses import dataclass
from io import BytesIO

import conllu
import numpy as np
import requests
import torch
from datasets import load_dataset
from mosestokenizer import MosesTokenizer
from tqdm.auto import tqdm
from transformers import HfArgumentParser

from wtpsplit.evaluation import preprocess_sentence
from wtpsplit.utils import Constants

UD_TREEBANK_PATH = "../data/ud-treebanks-v2.13"  # source: https://universaldependencies.org/#download

ERSATZ_DATA_PATH = "../data/ersatz-test-suite/segmented"  # source: https://github.com/rewicks/ersatz-test-suite

ERSATZ_TEST_DATASETS = {
    "ar": "iwsltt2017.ar",
    "cs": "wmt20.cs-en.cs",
    "de": "wmt20.de-en.de",
    "en": "wsj.03-06.en",
    "es": "wmt13.es-en.es",
    "et": "wmt18.et-en.et",
    "fi": "wmt19.fi-en.fi",
    "fr": "wmt20.fr-de.fr",
    "gu": "wmt19.gu-en.gu",
    "hi": "wmt14.hi-en.hi",
    "iu": "wmt20.iu-en.iu",
    "ja": "wmt20.ja-en.ja",
    "kk": "wmt19.kk-en.kk",
    "km": "wmt20.km-en.km",
    "lt": "wmt19.lt-en.lt",
    "lv": "wmt17.lv-en.lv",
    "pl": "wmt20.pl-en.pl",
    "ps": "wmt20.ps-en.ps",
    "ro": "wmt16.ro-en.ro",
    "ru": "wmt20.ru-en.ru",
    "ta": "wmt20.ta-en.ta",
    "tr": "wmt18.tr-en.tr",
    "zh": "wmt20.zh-en.zh",
}
ERSATZ_TRAIN_DATASETS = {
    "ar": "news-commentary-v15.dev.ar",
    "cs": "wmt18.cs-en.cs",
    "de": "wmt19.de-en.de",
    "en": "merged.nc-wsj.en",
    "es": None,
    "et": "newscrawl.2019.dev.et",
    "fi": "wmt18.fi-en.fi",
    "fr": "wmt15.fr-en.fr",
    "gu": "newscrawl.2019.dev.gu",
    "hi": "newscrawl.2013.dev.hi",
    "iu": "nhi-3.0.iu",
    "ja": "newscrawl.2019.dev.ja",
    "kk": "newscrawl.2019.dev.kk",
    "km": "wikimatrix.dev.km",
    "lt": "newscrawl.2019.dev.lt",
    "lv": "newscrawl.2019.dev.lv",
    "pl": "newscrawl.2019.dev.pl",
    "ps": "wikimatrix.dev.ps",
    "ro": "newscrawl.2019.dev.ro",
    "ru": "wmt18.ru-en.ru",
    "ta": "newscrawl.2019.dev.ta",
    "tr": "wmt16.tr-en.tr",
    "zh": "wmt18.zh-en.zh",
}


punct_chars = set(Constants.PUNCTUATION_CHARS)


def corrupt_asr(sentences, lang):
    # first corruption scheme of SM in the SaT paper
    if sentences is None:
        return None

    separator = Constants.SEPARATORS.get(lang, " ")

    if separator == "":
        corrupted_sentences = [
            preprocess_sentence("".join([char for char in sentence if char not in punct_chars]).lower())
            for sentence in sentences
        ]
        return corrupted_sentences

    try:
        tokenizer = MosesTokenizer(lang)
    except:  # noqa
        corrupted_sentences = [
            preprocess_sentence("".join([char for char in sentence if char not in punct_chars]).lower())
            for sentence in sentences
        ]
        return corrupted_sentences

    tokenized_sentences = [tokenizer.tokenize(sentence) for sentence in sentences]
    corrupted_tokenized_sentences = [
        [token for token in tokens if token not in punct_chars] for tokens in tokenized_sentences
    ]

    corrupted_sentences = [
        preprocess_sentence(tokenizer.detokenize(corrupted_tokens).lower())
        for corrupted_tokens in corrupted_tokenized_sentences
    ]

    return corrupted_sentences


def corrupt_social_media(sentences, lang):
    # second corruption scheme of SM in the SaT paper
    if sentences is None:
        return None

    corrupted_sentences = []
    for sentence in sentences:
        if random.random() < 0.5:
            sentence = "".join([char for char in sentence if char not in punct_chars])
        if random.random() < 0.5:
            sentence = sentence.lower()

        for punct in punct_chars:
            count = 0
            while random.random() < 0.5:
                count += 1
            sentence = sentence.replace(punct, punct * count)

        sentence = preprocess_sentence(sentence)
        corrupted_sentences.append(sentence)

    return corrupted_sentences


@dataclass
class Args:
    output_file: str = "../data/preprocessed_training_data/all_data.pth"
    include_train_data: bool = True
    cache_dir: str = "../data/cache/"


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()

    eval_data = {lang_code: {"sentence": {}, "compound": {}} for lang_code in Constants.LANGINFO.index}

    # Ersatz data
    for lang_code in tqdm(Constants.LANGINFO.index):
        if lang_code in ERSATZ_TEST_DATASETS:
            eval_data[lang_code]["sentence"]["ersatz"] = {
                "meta": {
                    "train_data": (
                        [
                            preprocess_sentence(line)
                            for line in open(
                                os.path.join(
                                    ERSATZ_DATA_PATH,
                                    lang_code,
                                    ERSATZ_TRAIN_DATASETS[lang_code],
                                )
                            )
                        ]
                        if args.include_train_data and ERSATZ_TRAIN_DATASETS[lang_code] is not None
                        else None
                    ),
                },
                "data": [
                    preprocess_sentence(line)
                    for line in open(os.path.join(ERSATZ_DATA_PATH, lang_code, ERSATZ_TEST_DATASETS[lang_code]))
                ],
            }

            eval_data[lang_code]["sentence"]["ersatz-corrupted-asr"] = {
                "meta": {
                    "train_data": (
                        corrupt_asr(
                            (
                                eval_data[lang_code]["sentence"]["ersatz"]["meta"]["train_data"][:10000]
                                if eval_data[lang_code]["sentence"]["ersatz"]["meta"]["train_data"] is not None
                                else None
                            ),
                            lang_code,
                        )
                    )
                },
                "data": corrupt_asr(
                    eval_data[lang_code]["sentence"]["ersatz"]["data"][:10000],
                    lang_code,
                ),
            }

            eval_data[lang_code]["sentence"]["ersatz-corrupted-social-media"] = {
                "meta": {
                    "train_data": (
                        corrupt_social_media(
                            (
                                eval_data[lang_code]["sentence"]["ersatz"]["meta"]["train_data"][:10000]
                                if eval_data[lang_code]["sentence"]["ersatz"]["meta"]["train_data"] is not None
                                else None
                            ),
                            lang_code,
                        )
                    )
                },
                "data": corrupt_social_media(
                    eval_data[lang_code]["sentence"]["ersatz"]["data"][:10000],
                    lang_code,
                ),
            }

    # UD + OPUS100 sentences + TED + NLLB
    for lang_code in tqdm(Constants.LANGINFO.index):
        opus_dset_name = Constants.LANGINFO.loc[lang_code, "opus100"]

        if opus_dset_name not in (np.nan, None):
            other_lang_code = set(opus_dset_name.split("-")) - {lang_code}
            assert len(other_lang_code) == 1
            other_lang_code = other_lang_code.pop()

            dset_args = ["opus100", opus_dset_name]

            try:
                opus100_sentences = [
                    preprocess_sentence(sample[lang_code])
                    for sample in load_dataset(*dset_args, split="test", cache_dir=args.cache_dir)["translation"]
                    if sample[lang_code].strip() != sample[other_lang_code].strip()
                ]
                try:
                    opus100_train_sentences = [
                        preprocess_sentence(sample[lang_code])
                        for sample in load_dataset(*dset_args, split="train", cache_dir=args.cache_dir)["translation"]
                        if sample[lang_code].strip() != sample[other_lang_code].strip()
                    ]
                except ValueError:
                    opus100_train_sentences = None
            except ValueError:
                opus100_sentences = [
                    preprocess_sentence(sample[lang_code])
                    for sample in load_dataset(*dset_args, split="train", cache_dir=args.cache_dir)["translation"]
                    if sample[lang_code].strip() != sample[other_lang_code].strip()
                ]
                opus100_train_sentences = None

            opus100_train_sentences = opus100_train_sentences[:10000] if opus100_train_sentences is not None else None

            eval_data[lang_code]["sentence"]["opus100"] = {
                "meta": {"train_data": (opus100_train_sentences if args.include_train_data else None)},
                "data": opus100_sentences,
            }

            eval_data[lang_code]["sentence"]["opus100-corrupted-asr"] = {
                "meta": {
                    "train_data": (
                        corrupt_asr(
                            (opus100_train_sentences),
                            lang_code,
                        )
                        if args.include_train_data
                        else None
                    )
                },
                "data": corrupt_asr(opus100_sentences[:10000], lang_code),
            }

            eval_data[lang_code]["sentence"]["opus100-corrupted-social-media"] = {
                "meta": {
                    "train_data": (
                        corrupt_social_media(
                            (opus100_train_sentences),
                            lang_code,
                        )
                        if args.include_train_data
                        else None
                    )
                },
                "data": corrupt_social_media(opus100_sentences[:10000], lang_code),
            }

        if Constants.LANGINFO.loc[lang_code, "ud"] not in (np.nan, None):
            ud_data = conllu.parse(
                open(
                    glob.glob(
                        os.path.join(
                            UD_TREEBANK_PATH,
                            Constants.LANGINFO.loc[lang_code, "ud"],
                            "*-ud-test.conllu",
                        )
                    )[0]
                ).read()
            )

            try:
                ud_train_data = conllu.parse(
                    open(
                        glob.glob(
                            os.path.join(
                                UD_TREEBANK_PATH,
                                Constants.LANGINFO.loc[lang_code, "ud"],
                                "*-ud-train.conllu",
                            )
                        )[0]
                    ).read()
                )
                ud_train_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_train_data][
                    :10000
                ]
            except IndexError:
                ud_train_sentences = None

            ud_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_data]

            eval_data[lang_code]["sentence"]["ud"] = {
                "meta": {"train_data": (ud_train_sentences if args.include_train_data else None)},
                "data": ud_sentences,
            }

            eval_data[lang_code]["sentence"]["ud-corrupted-asr"] = {
                "meta": {
                    "train_data": (corrupt_asr(ud_train_sentences, lang_code) if args.include_train_data else None)
                },
                "data": corrupt_asr(ud_sentences, lang_code),
            }

            eval_data[lang_code]["sentence"]["ud-corrupted-social-media"] = {
                "meta": {
                    "train_data": (
                        corrupt_social_media(ud_train_sentences, lang_code) if args.include_train_data else None
                    )
                },
                "data": corrupt_social_media(ud_sentences, lang_code),
            }

        # TED 2020
        url = f"https://object.pouta.csc.fi/OPUS-TED2020/v1/mono/{lang_code}.txt.gz"
        res = requests.get(url)

        if res.status_code == 200:
            with gzip.open(BytesIO(res.content), "rt", encoding="utf-8") as f:
                sentences = f.read().splitlines()

            sentences = sentences[:20000]

            sentences = [preprocess_sentence(sentence) for sentence in sentences]

            train_sentences = sentences[: len(sentences) // 2]
            test_sentences = sentences[len(sentences) // 2 :]

            eval_data[lang_code]["sentence"]["ted2020-corrupted-asr"] = {
                "meta": {"train_data": (corrupt_asr(train_sentences, lang_code) if args.include_train_data else None)},
                "data": corrupt_asr(test_sentences, lang_code),
            }

            eval_data[lang_code]["sentence"]["ted2020-corrupted-social-media"] = {
                "meta": {
                    "train_data": (
                        corrupt_social_media(train_sentences, lang_code) if args.include_train_data else None
                    )
                },
                "data": corrupt_social_media(test_sentences, lang_code),
            }

        else:
            print(f"Failed to download TED2020 data for {lang_code}")

    for lang_code in ["ceb", "jv", "mn", "yo"]:
        url = f"https://object.pouta.csc.fi/OPUS-NLLB/v1/mono/{lang_code}.txt.gz"
        res = requests.get(url)

        if res.status_code == 200:
            with gzip.open(BytesIO(res.content), "rt", encoding="utf-8") as f:
                sentences = f.read().splitlines()

            random.shuffle(sentences)  # because they come alphabetically sorted

            sentences = sentences[:20000]

            sentences = [preprocess_sentence(sentence) for sentence in sentences]

        else:
            raise Exception

        train_sentences = sentences[: len(sentences) // 2]
        test_sentences = sentences[len(sentences) // 2 :]

        eval_data[lang_code]["sentence"]["nllb"] = {
            "meta": {"train_data": (train_sentences if args.include_train_data else None)},
            "data": test_sentences,
        }

        eval_data[lang_code]["sentence"]["nllb-corrupted-asr"] = {
            "meta": {"train_data": (corrupt_asr(train_sentences, lang_code) if args.include_train_data else None)},
            "data": corrupt_asr(test_sentences, lang_code),
        }

        eval_data[lang_code]["sentence"]["nllb-corrupted-social-media"] = {
            "meta": {
                "train_data": (corrupt_social_media(train_sentences, lang_code) if args.include_train_data else None)
            },
            "data": corrupt_social_media(test_sentences, lang_code),
        }

    # UD Code-Switching Corpora

    # UD_Turkish_German-SAGT

    ud_data = conllu.parse(
        open(
            glob.glob(
                os.path.join(
                    UD_TREEBANK_PATH,
                    "UD_Turkish_German-SAGT",
                    "*-ud-test.conllu",
                )
            )[0]
        ).read()
    )

    ud_train_data = conllu.parse(
        open(
            glob.glob(
                os.path.join(
                    UD_TREEBANK_PATH,
                    "UD_Turkish_German-SAGT",
                    "*-ud-train.conllu",
                )
            )[0]
        ).read()
    )

    ud_train_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_train_data]
    ud_test_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_data]

    eval_data["tr-de"] = {}
    eval_data["tr-de"]["sentence"] = {}

    eval_data["tr-de"]["sentence"]["code-switching"] = {
        "meta": {"train_data": ud_train_sentences},
        "data": ud_test_sentences,
    }

    eval_data["tr-de"]["sentence"]["code-switching-corrupted-asr"] = {
        "meta": {"train_data": corrupt_asr(ud_train_sentences, "en")},
        "data": corrupt_asr(ud_test_sentences, "en"),
    }

    eval_data["tr-de"]["sentence"]["code-switching-corrupted-social-media"] = {
        "meta": {"train_data": corrupt_social_media(ud_train_sentences, "en")},
        "data": corrupt_social_media(ud_test_sentences, "en"),
    }

    # UD_Spanish_English-Miami

    ud_data = conllu.parse(
        open(
            glob.glob(
                os.path.join(
                    UD_TREEBANK_PATH,
                    "UD_Spanish_English-Miami",
                    "*-ud-test.conllu",
                )
            )[0]
        ).read()
    )

    ud_sentences = [preprocess_sentence(sentence.metadata["text"]) for sentence in ud_data]
    ud_train_sentences = ud_sentences[len(ud_sentences) // 2 :]
    ud_test_sentences = ud_sentences[: len(ud_sentences) // 2]

    eval_data["es-en"] = {}
    eval_data["es-en"]["sentence"] = {}

    eval_data["es-en"]["sentence"]["code-switching"] = {
        "meta": {"train_data": ud_train_sentences},
        "data": ud_test_sentences,
    }

    eval_data["es-en"]["sentence"]["code-switching-corrupted-asr"] = {
        "meta": {"train_data": corrupt_asr(ud_train_sentences, "es")},
        "data": corrupt_asr(ud_test_sentences, "es"),
    }

    eval_data["es-en"]["sentence"]["code-switching-corrupted-social-media"] = {
        "meta": {"train_data": corrupt_social_media(ud_train_sentences, "es")},
        "data": corrupt_social_media(ud_test_sentences, "es"),
    }

    # Short sequences

    # serbian

    serbian_train_data = conllu.parse(open("../data/short-sequences/serbian/reldi-normtagner-sr-train.conllu").read())

    serbian_train_tweets = []
    tweet_sentences = []
    for sentence in serbian_train_data:
        if "newdoc id" in sentence.metadata:
            if tweet_sentences:
                serbian_train_tweets.append(tweet_sentences)
            tweet_sentences = []
        tweet_sentences.append(preprocess_sentence(sentence.metadata["text"]))

    if tweet_sentences:
        serbian_train_tweets.append(tweet_sentences)

    serbian_test_data = conllu.parse(open("../data/short-sequences/serbian/reldi-normtagner-sr-test.conllu").read())

    serbian_test_tweets = []
    tweet_sentences = []
    for sentence in serbian_test_data:
        if "newdoc id" in sentence.metadata:
            if tweet_sentences:
                serbian_test_tweets.append(tweet_sentences)
            tweet_sentences = []
        tweet_sentences.append(preprocess_sentence(sentence.metadata["text"]))

    if tweet_sentences:
        serbian_test_tweets.append(tweet_sentences)

    serbian_train_tweets = [tweet for tweet in serbian_train_tweets if len(tweet) > 1]
    serbian_test_tweets = [tweet for tweet in serbian_test_tweets if len(tweet) > 1]

    eval_data["sr"]["sentence"]["short-sequences"] = {
        "meta": {"train_data": serbian_train_tweets},
        "data": serbian_test_tweets,
    }

    eval_data["sr"]["sentence"]["short-sequences-corrupted-asr"] = {
        "meta": {"train_data": [corrupt_asr(s, "sr") for s in serbian_train_tweets]},
        "data": [corrupt_asr(s, "sr") for s in serbian_test_tweets],
    }

    eval_data["sr"]["sentence"]["short-sequences-corrupted-social-media"] = {
        "meta": {"train_data": [corrupt_social_media(s, "sr") for s in serbian_train_tweets]},
        "data": [corrupt_social_media(s, "sr") for s in serbian_test_tweets],
    }

    # slovenian

    slovenian_data = conllu.parse(
        open("../data/short-sequences/slovenian/Janes-Tag.3.0.CoNLL-U/janes-rsdo.ud.connlu").read()
    )

    slovenian_tweets = []
    tweet_sentences = []
    for sentence in slovenian_data:
        if "newdoc id" in sentence.metadata:
            if tweet_sentences:
                slovenian_tweets.append(tweet_sentences)
            tweet_sentences = []
        tweet_sentences.append(preprocess_sentence(sentence.metadata["text"]))

    if tweet_sentences:
        slovenian_tweets.append(tweet_sentences)

    random.shuffle(slovenian_tweets)

    # keep only if more than one sentence in a tweet
    slovenian_tweets = [tweet for tweet in slovenian_tweets if len(tweet) > 1]

    slovenian_train_tweeets = slovenian_tweets[: len(slovenian_tweets) // 2]
    slovenian_test_tweets = slovenian_tweets[len(slovenian_tweets) // 2 :]

    eval_data["sl"]["sentence"]["short-sequences"] = {
        "meta": {"train_data": slovenian_train_tweeets},
        "data": slovenian_test_tweets,
    }

    eval_data["sl"]["sentence"]["short-sequences-corrupted-asr"] = {
        "meta": {"train_data": [corrupt_asr(s, "sl") for s in slovenian_train_tweeets]},
        "data": [corrupt_asr(s, "sl") for s in slovenian_test_tweets],
    }

    eval_data["sl"]["sentence"]["short-sequences-corrupted-social-media"] = {
        "meta": {"train_data": [corrupt_social_media(s, "sl") for s in slovenian_train_tweeets]},
        "data": [corrupt_social_media(s, "sl") for s in slovenian_test_tweets],
    }

    # LEGAL
    langs = ["de", "en", "es", "fr", "it"]

    all_subset_data = {
        lang: {
            "laws": {"train": [], "test": []},
            "judgements": {"train": [], "test": []},
        }
        for lang in langs
    }

    for lang in tqdm(langs, desc="Legal data"):
        data_dir = f"../data/MultiLegalSBD/data/{lang}/gold/"

        all_files = glob.glob(f"{data_dir}/*_test.jsonl")
        subsets = [file.split("/")[-1].split("_test.jsonl")[0] for file in all_files]

        for subset in subsets:
            if subset == "Constitution":
                continue

            train_data = []

            with open(
                f"../data/MultiLegalSBD/data/{lang}/gold/{subset}_train.jsonl",
                "r",
                encoding="utf-8",
            ) as f:
                for line in f:
                    train_data.append(json.loads(line))

            train_subset_sentences = []
            for doc in train_data:
                doc_sentences = []
                text = doc["text"]
                for span in doc["spans"]:
                    sentence = text[span["start"] : span["end"]]
                    doc_sentences.append(preprocess_sentence(sentence))
                train_subset_sentences.append(doc_sentences)

            test_data = []

            test_data_file = f"../data/MultiLegalSBD/data/{lang}/gold/{subset}_test.jsonl"

            with open(
                test_data_file,
                "r",
                encoding="utf-8",
            ) as f:
                for line in f:
                    test_data.append(json.loads(line))

            test_subset_sentences = []
            for doc in test_data:
                doc_sentences = []
                text = doc["text"]
                for span in doc["spans"]:
                    sentence = text[span["start"] : span["end"]]
                    doc_sentences.append(preprocess_sentence(sentence))
                test_subset_sentences.append(doc_sentences)

            eval_data[lang]["sentence"][f"legal-{subset}"] = {
                "meta": {"train_data": train_subset_sentences},
                "data": test_subset_sentences,
            }

            eval_data[lang]["sentence"][f"legal-{subset}-corrupted-asr"] = {
                "meta": {"train_data": [corrupt_asr(s, lang) for s in train_subset_sentences]},
                "data": [corrupt_asr(s, lang) for s in test_subset_sentences],
            }

            eval_data[lang]["sentence"][f"legal-{subset}-corrupted-social-media"] = {
                "meta": {"train_data": [corrupt_social_media(s, lang) for s in train_subset_sentences]},
                "data": [corrupt_social_media(s, lang) for s in test_subset_sentences],
            }

            subsets2set = {
                "CD_jug": "judgements",
                "gesCode": "laws",
                "CD_multi_legal": "judgements",
                "CD_wipolex": "judgements",
                "CivilCode": "laws",
                "CriminalCode": "laws",
                "CD_swiss_judgement": "judgements",
            }

            if lang != "en":
                set = subsets2set[subset]
            else:
                set = "judgements"

            all_subset_data[lang][set]["train"].extend(train_subset_sentences)
            all_subset_data[lang][set]["test"].extend(test_subset_sentences)

        # constitution

        if lang in ["de", "en"]:
            continue

        test_data = []
        test_data_file = f"../data/MultiLegalSBD/data/{lang}/gold/Constitution.jsonl"
        with open(
            test_data_file,
            "r",
            encoding="utf-8",
        ) as f:
            for line in f:
                test_data.append(json.loads(line))

        test_subset_sentences = []
        for doc in test_data:
            doc_sentences = []
            text = doc["text"]
            for span in doc["spans"]:
                sentence = text[span["start"] : span["end"]]
                doc_sentences.append(preprocess_sentence(sentence))
            test_subset_sentences.append(doc_sentences)

        eval_data[lang]["sentence"]["legal-constitution"] = {
            "meta": {"train_data": None},
            "data": test_subset_sentences,
        }

        eval_data[lang]["sentence"]["legal-constitution-corrupted-asr"] = {
            "meta": {"train_data": None},
            "data": [corrupt_asr(s, lang) for s in test_subset_sentences],
        }

        eval_data[lang]["sentence"]["legal-constitution-corrupted-social-media"] = {
            "meta": {"train_data": None},
            "data": [corrupt_social_media(s, lang) for s in test_subset_sentences],
        }

        all_subset_data[lang]["laws"]["test"].extend(test_subset_sentences)

    for lang in all_subset_data:
        for set in ["laws", "judgements"]:
            eval_data[lang]["sentence"][f"legal-all-{set}"] = {
                "meta": {"train_data": all_subset_data[lang][set]["train"]},
                "data": all_subset_data[lang][set]["test"],
            }

            eval_data[lang]["sentence"][f"legal-all-{set}-corrupted-asr"] = {
                "meta": {"train_data": [corrupt_asr(s, lang) for s in all_subset_data[lang][set]["train"]]},
                "data": [corrupt_asr(s, lang) for s in all_subset_data[lang][set]["test"]],
            }

            eval_data[lang]["sentence"][f"legal-all-{set}-corrupted-social-media"] = {
                "meta": {"train_data": [corrupt_social_media(s, lang) for s in all_subset_data[lang][set]["train"]]},
                "data": [corrupt_social_media(s, lang) for s in all_subset_data[lang][set]["test"]],
            }

    torch.save(eval_data, args.output_file.replace(".pth", "-all.pth"))



import json
import os
import sys

from sklearn.metrics import classification_report

from wtpsplit.utils import Constants


def evaluate_subtask1(splits, langs, prediction_dir: str, supervisions, include_n_documents) -> None:
    """
    Mirrors the original SEPP-NLG 2021 Shared Task evaluation function
    https://sites.google.com/view/sentence-segmentation
    """

    results = {}
    avg_holder = {}
    for supervision in supervisions:
        avg_holder[supervision] = 0
    for lang_code in langs:
        results[lang_code] = {}
        for split in splits:
            results[lang_code][split] = {}
            for supervision in supervisions:
                print(lang_code, split, supervision)
                relevant_dir = Constants.ROOT_DIR.parent / "data/sepp_nlg_2021_data" / lang_code / split

                all_gt_labels, all_predicted_labels = [], []
                fnames = sorted([f for f in relevant_dir.glob("*.tsv") if f.is_file()])[:include_n_documents]
                gt_tsv_files = [
                    fname
                    for fname in fnames
                    if str(fname).startswith(str(relevant_dir)) and str(fname).endswith(".tsv")
                ]

                for _, gt_tsv_file in enumerate(gt_tsv_files, 0):
                    basename = os.path.basename(gt_tsv_file)

                    with open(gt_tsv_file, encoding="utf-8") as f:
                        lines = f.read().strip().split("\n")
                        rows = [line.split("\t") for line in lines]
                        gt_labels = [row[1] for row in rows]

                    with open(
                        os.path.join(
                            Constants.CACHE_DIR, "ted2020", prediction_dir, lang_code, split, supervision, basename
                        ),
                        "r",
                        encoding="utf8",
                    ) as f:
                        lines = f.read().strip().split("\n")
                        rows = [line.split("\t") for line in lines]
                        pred_labels = [row[1] for row in rows]

                    assert (
                        len(gt_labels) == len(pred_labels)
                    ), f"unequal no. of labels for files {gt_tsv_file} and {os.path.join(prediction_dir,  lang_code, split, basename)}"
                    all_gt_labels.extend(gt_labels)
                    all_predicted_labels.extend(pred_labels)

                eval_result = classification_report(all_gt_labels, all_predicted_labels, output_dict=True)
                print(eval_result["1"]["f1-score"])
                avg_holder[supervision] += eval_result["1"]["f1-score"]
                results[lang_code][split][supervision] = eval_result
    results["avg"] = {}
    for supervision in supervisions:
        avg_holder[supervision] /= len(langs)
        results["avg"][supervision] = avg_holder[supervision]
    print(avg_holder)
    json.dump(
        results,
        open(
            Constants.CACHE_DIR / "ted2020" / f"{prediction_dir}_TED.json",
            "w",
        ),
        indent=4,
    )
    print(Constants.CACHE_DIR / "ted2020" / f"{prediction_dir}_TED.json")
    return results


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Evaluate subtask 1 of SEPP-NLG 2021")
    parser.add_argument(
        "--include_languages",
        help="target language ('en', 'de', 'fr', 'it'; i.e. one of the subfolders in the zip file's main folder)",
        default=["fr", "de", "en", "it"],
        nargs="+",
    )
    parser.add_argument(
        "--splits",
        help="split to be evaluated (usually 'dev', 'test'), subfolder of 'lang'",
        default=["test", "surprise_test"],
        nargs="+",
    )
    parser.add_argument(
        "--prediction_dir",
        help="path to folder containing the prediction TSVs (language and test set folder names are appended automatically)",
    )
    parser.add_argument(
        "--supervision",
        help="u, t, punct",
        default=["u", "t", "punct"],
        nargs="+",
    )
    parser.add_argument("--include_n_documents", default=sys.maxsize)
    args = parser.parse_args()
    results = evaluate_subtask1(args.splits, args.include_languages, args.prediction_dir, args.supervision)



import re
import torch


def remove_emojis_and_special_chars(text):
    emoji_pattern = re.compile(
        "["
        "\U0001f600-\U0001f64f"  # emoticons
        "\U0001f300-\U0001f5ff"  # symbols & pictographs
        "\U0001f680-\U0001f6ff"  # transport & map symbols
        "\U0001f700-\U0001f77f"  # alchemical symbols
        "\U0001f780-\U0001f7ff"  # Geometric Shapes Extended
        "\U0001f800-\U0001f8ff"  # Supplemental Arrows-C
        "\U0001f900-\U0001f9ff"  # Supplemental Symbols and Pictographs
        "\U0001fa00-\U0001fa6f"  # Chess Symbols
        "\U0001fa70-\U0001faff"  # Symbols and Pictographs Extended-A
        "\U00002702-\U000027b0"  # Dingbats
        "\U000024c2-\U0001f251"
        "]+",
        flags=re.UNICODE,
    )
    text = emoji_pattern.sub(r"", text)
    text = re.sub(r"[:;=Xx][\-oO\']*[\)\(\[\]DdPp3><\|\\\/]", "", text)
    return text


def transform_data(data):
    def pair_sentences(sequences):
        paired_sequences = []
        for sequence in sequences:
            processed_sequence = []
            for sentence in sequence:
                words = sentence.strip().split()
                filtered_words = [
                    remove_emojis_and_special_chars(word)
                    for word in words
                    if not (word.startswith("http") or word.startswith("#") or word.startswith("@"))
                ]
                cleaned_sentence = " ".join(filtered_words)  # fine for our langs.
                if cleaned_sentence and len(cleaned_sentence.split()) > 0:
                    processed_sequence.append(cleaned_sentence.strip())
            if processed_sequence and len(processed_sequence) < 6:
                paired_sequences.append(processed_sequence)
        return paired_sequences

    transformed_data = {}
    for lang_code, lang_data in data.items():
        transformed_data[lang_code] = {}
        for content_type, datasets in lang_data.items():
            if content_type != "sentence":
                continue
            transformed_data[lang_code] = {}
            transformed_data[lang_code][content_type] = {}
            for dataset_name, content in datasets.items():
                if "short" not in dataset_name:
                    continue
                transformed_data[lang_code][content_type][dataset_name] = {
                    "meta": {"train_data": pair_sentences(content["meta"]["train_data"])},
                    "data": pair_sentences(content["data"]),
                }

    return transformed_data


data = torch.load("data/all_data.pth")

transformed_data = transform_data(data)
torch.save(transformed_data, "data/all_data_tweets_cleaned.pth")



from pathlib import Path

import torch
from mosestokenizer import MosesTokenizer
from tqdm.auto import tqdm

from wtpsplit.utils import Constants


def process_tsv(file_path, detokenizer):
    sentences = []
    current_sentence = []

    with open(file_path, "r", encoding="utf-8") as file:
        for line in file:
            parts = line.strip().split("\t")
            if len(parts) < 2:
                continue
            word, boundary = parts[0], parts[1]
            current_sentence.append(word)
            if boundary == "1":
                detokenized_sentence = detokenizer.detokenize(current_sentence)
                # detokenized_sentence = ' '.join(current_sentence)
                sentences.append(detokenized_sentence)
                current_sentence = []

    return sentences


def build_data_dictionary(root_dir):
    data_dict = {}

    for lang in tqdm(["fr", "de", "en", "it"]):
        detokenizer = MosesTokenizer(lang)
        data_dict[lang] = {"sentence": {}}
        for dataset in ["test", "surprise_test"]:
            data_dict[lang]["sentence"][dataset] = {"meta": {"train_data": []}, "data": []}

        train_data_path = Path(root_dir) / lang / "train"
        train_files = sorted([f for f in train_data_path.glob("*.tsv") if f.is_file()])
        all_train_sentences = []
        for file_path in tqdm(train_files, desc=f"{lang} train"):
            train_sentences = process_tsv(file_path, detokenizer)
            all_train_sentences.append(train_sentences)

        # use train data for both test sets (same training data)
        for dataset in ["test", "surprise_test"]:
            data_dict[lang]["sentence"][dataset]["meta"]["train_data"] = all_train_sentences

        # test + surprise_test data
        for dataset in ["test", "surprise_test"]:
            test_data_path = Path(root_dir) / lang / dataset
            test_files = sorted([f for f in test_data_path.glob("*.tsv") if f.is_file()])
            all_test_sentences = []
            for file_path in tqdm(test_files, desc=f"{lang} {dataset}"):
                test_sentences = process_tsv(file_path, detokenizer)
                all_test_sentences.append(test_sentences)
            data_dict[lang]["sentence"][dataset]["data"] = all_test_sentences

    return data_dict


# this must be downloaded first from the linked drive here
# https://sites.google.com/view/sentence-segmentation
root_dir = Constants.ROOT_DIR.parent / "data/sepp_nlg_2021_data"
data = build_data_dictionary(root_dir)

torch.save(data, Constants.ROOT_DIR.parent / "data" / "ted2020_join.pth")



import json
from dataclasses import dataclass
from typing import List

import torch
from tqdm import tqdm
from transformers import HfArgumentParser

from wtpsplit.evaluation import (
    LanguageError,
    ersatz_sentencize,
    evaluate_sentences,
    punkt_sentencize,
    pysbd_sentencize,
    spacy_dp_sentencize,
    spacy_sent_sentencize,
)
from wtpsplit.utils import Constants


def split_language_data(eval_data):
    # used if 2 language codes given (i.e., code-switching)
    new_eval_data = {}

    for lang_code, lang_data in eval_data.items():
        if "-" in lang_code:
            lang1, lang2 = lang_code.split("-")
            new_lang1 = f"{lang_code}_{lang1.upper()}"
            new_lang2 = f"{lang_code}_{lang2.upper()}"

            new_eval_data[new_lang1] = lang_data
            new_eval_data[new_lang2] = lang_data
        else:
            new_eval_data[lang_code] = lang_data

    return new_eval_data


@dataclass
class Args:
    eval_data_path: str = "data/all_data.pth"
    include_langs: List[str] = None
    exclude_every_k: int = 10


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()

    eval_data = torch.load(args.eval_data_path)
    eval_data = split_language_data(eval_data)
    results = {}
    indices = {}

    for lang, lang_data in tqdm(eval_data.items()):
        if args.include_langs is not None and lang not in args.include_langs:
            continue

        results[lang] = {}
        indices[lang] = {}

        for dataset_name, dataset in lang_data["sentence"].items():
            if "nllb" in dataset_name:
                continue
            if not dataset["data"]:
                continue
            results[lang][dataset_name] = {}
            indices[lang][dataset_name] = {}

            if "-" in lang:
                # code-switched data: eval 2x
                lang_code = lang.split("_")[1].lower()
            else:
                lang_code = lang

            for f, name in [
                (punkt_sentencize, "punkt"),
                (spacy_dp_sentencize, "spacy_dp"),
                (spacy_sent_sentencize, "spacy_sent"),
                (pysbd_sentencize, "pysbd"),
                (ersatz_sentencize, "ersatz"),
            ]:
                print(f"Running {name} on {dataset_name} in {lang_code}...")
                indices[lang][dataset_name][name] = {}
                if "lyrics" in dataset_name or "short" in dataset_name:
                    exclude_every_k = 0
                else:
                    exclude_every_k = args.exclude_every_k
                try:
                    if isinstance(dataset["data"][0], list):
                        all_sentences = dataset["data"]
                        metrics = []
                        for i, sentences in enumerate(all_sentences):
                            text = Constants.SEPARATORS[lang_code].join(sentences)
                            doc_metrics = {}
                            doc_metrics = evaluate_sentences(
                                lang_code,
                                sentences,
                                f(lang_code, text),
                                return_indices=True,
                                exclude_every_k=exclude_every_k,
                            )
                            f1 = doc_metrics[0]
                            doc_metrics = doc_metrics[1]
                            doc_metrics["f1"] = f1
                            doc_metrics["length"] = [doc_metrics["length"]]
                            metrics.append(doc_metrics)
                        avg_results = {}
                        concat_indices = {}
                        for doc in metrics:
                            for key, value in doc.items():
                                if not isinstance(value, list):
                                    # numeric
                                    if key not in avg_results:
                                        avg_results[key] = []

                                    avg_results[key].append(value)
                                elif isinstance(value, list):
                                    # concat
                                    if key not in concat_indices:
                                        concat_indices[key] = []
                                    if key == "length":
                                        concat_indices[key].append(value[0])
                                    else:
                                        concat_indices[key].append(value)

                        # avg
                        for key in list(avg_results):
                            if avg_results[key]:
                                avg_results[key] = sum(avg_results[key]) / len(avg_results[key])

                        # Store the results and indices
                        results[lang][dataset_name][name] = avg_results
                        indices[lang][dataset_name][name] = concat_indices
                    else:
                        # sentences = [preprocess_sentence(s) for s in dataset["data"]]
                        sentences = dataset["data"]
                        text = Constants.SEPARATORS[lang_code].join(sentences)

                        metrics = evaluate_sentences(
                            lang_code,
                            sentences,
                            f(lang_code, text),
                            return_indices=True,
                            exclude_every_k=exclude_every_k,
                        )
                        f1 = metrics[0]
                        metrics = metrics[1]
                        metrics["f1"] = f1
                        print(f1)
                        indices[lang][dataset_name][name]["true_indices"] = [metrics.pop("true_indices")]
                        indices[lang][dataset_name][name]["predicted_indices"] = [metrics.pop("predicted_indices")]
                        indices[lang][dataset_name][name]["length"] = [metrics.pop("length")]
                        results[lang][dataset_name][name] = metrics
                except LanguageError as e:
                    print("Language not supported for", name, e)
                    results[lang][dataset_name][name] = None

    json.dump(results, open(Constants.CACHE_DIR / "intrinsic_baselines.json", "w"), indent=4, default=int)
    json.dump(indices, open(Constants.CACHE_DIR / "intrinsic_baselines_IDX.json", "w"), indent=4, default=int)
    print(Constants.CACHE_DIR / "intrinsic_baselines.json")



import math
import sys
import logging

import numpy as np
from tqdm.auto import tqdm
from transformers import AutoTokenizer

from wtpsplit.utils import Constants, hash_encode

logger = logging.getLogger(__name__)


class BertCharORTWrapper:
    def __init__(self, config, ort_session):
        self.config = config
        self.ort_session = ort_session

    def __getattr__(self, name):
        assert hasattr(self, "ort_session")
        return getattr(self.ort_session, name)

    def __call__(self, hashed_ids, attention_mask):
        logits = self.ort_session.run(
            ["logits"],
            {
                "attention_mask": attention_mask.astype(np.float16),  # ORT expects fp16 mask
                "hashed_ids": hashed_ids,
            },
        )[0]

        return {"logits": logits}


class SaTORTWrapper:
    def __init__(self, config, ort_session):
        self.config = config
        self.ort_session = ort_session

    def __getattr__(self, name):
        assert hasattr(self, "ort_session")
        return getattr(self.ort_session, name)

    def __call__(self, input_ids, attention_mask):
        logits = self.ort_session.run(
            ["logits"],
            {
                "attention_mask": attention_mask.astype(np.float16),
                "input_ids": input_ids.astype(np.int64),
            },
        )[0]

        return {"logits": logits}


class PyTorchWrapper:
    def __init__(self, model):
        self.model = model
        self.config = model.config

    def __getattr__(self, name):
        assert hasattr(self, "model")
        return getattr(self.model, name)

    def __call__(self, attention_mask, hashed_ids=None, language_ids=None, input_ids=None):
        try:
            import torch
        except ImportError:
            raise ImportError("`torch` must be installed to use PyTorch models!")

        with torch.no_grad():
            logits = (
                self.model(
                    input_ids=torch.from_numpy(input_ids).to(self.model.device) if input_ids is not None else None,
                    hashed_ids=torch.from_numpy(hashed_ids).to(self.model.device) if hashed_ids is not None else None,
                    attention_mask=torch.from_numpy(attention_mask).to(self.model.device),
                    language_ids=(
                        torch.from_numpy(language_ids).to(self.model.device) if language_ids is not None else None
                    ),
                )["logits"]
                .cpu()
                .numpy()
            )

        return {"logits": logits}


def extract(
    batch_of_texts,
    model,
    stride,
    max_block_size,
    batch_size,
    lang_code=None,
    pad_last_batch=False,
    verbose=False,
    tokenizer=None,
):
    """
    Computes logits for the given batch of texts by:
        1. slicing the texts into chunks of size `block_size`.
        2. passing every chunk through the model forward.
        3. stitching predictings back together by averaging chunk logits.

    ad 1.: text is sliced into partially overlapping chunks by moving forward by a `stride` parameter (think conv1d).
    """
    if "xlm" in model.config.model_type:
        use_subwords = True
        if not tokenizer:
            tokenizer = AutoTokenizer.from_pretrained(
                "facebookAI/xlm-roberta-base",
            )
        # tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})
        tokens = tokenizer(batch_of_texts, return_offsets_mapping=True, verbose=False, add_special_tokens=False)
        # remove CLS and SEP tokens, they are added later anyhow
        # batch_of_texts = [text[1:-1] for text in tokens["input_ids"]]
        batch_of_texts = tokens["input_ids"]
        # offset_mapping = [offset[1:-1] for offset in tokens["offset_mapping"]]
        offset_mapping = tokens["offset_mapping"]
        cls_token_id = tokenizer.cls_token_id
        sep_token_id = tokenizer.sep_token_id
        pad_token_id = tokenizer.pad_token_id
    else:
        pad_token_id = 0
        use_subwords = False

    text_lengths = [len(text) for text in batch_of_texts]
    # reduce block size if possible
    block_size = min(max_block_size, max(text_lengths))
    if use_subwords and block_size > 510:
        overflow_length = block_size - 510
        block_size -= overflow_length  # account for CLS and SEP tokens

    # make sure block_size is a multiple of downsampling rate
    downsampling_rate = getattr(model.config, "downsampling_rate", 1)
    block_size = math.ceil(block_size / downsampling_rate) * downsampling_rate

    # total number of forward passes
    num_chunks = sum(math.ceil(max(length - block_size, 0) / stride) + 1 for length in text_lengths)

    # preallocate a buffer for all input hashes & attention masks
    if not use_subwords:
        input_hashes = np.zeros((num_chunks, block_size, model.config.num_hash_functions), dtype=np.int64)
        attention_mask = np.zeros((num_chunks, block_size), dtype=np.float32)
    else:
        input_ids = np.zeros((num_chunks, block_size + 2), dtype=np.int64)
        attention_mask = np.zeros((num_chunks, block_size + 2), dtype=np.float32)

    # locs keep track of the location of every chunk with a 3-tuple (text_idx, char_start, char_end) that indexes
    # back into the batch_of_texts
    locs = np.zeros((num_chunks, 3), dtype=np.int32)

    if not use_subwords:
        # this is equivalent to (but faster than) np.array([ord(c) for c in "".join(batch_of_texts)])
        codec = "utf-32-le" if sys.byteorder == "little" else "utf-32-be"
        ordinals = np.frombuffer(bytearray("".join(batch_of_texts), encoding=codec), dtype=np.int32)
        # hash encode all ids
        flat_hashed_ids = hash_encode(
            ordinals, num_hashes=model.config.num_hash_functions, num_buckets=model.config.num_hash_buckets
        )
    # note that ordinals and flat_hashed_ids have the same length
    offset = 0
    current_chunk = 0

    # create chunks
    for i in range(len(batch_of_texts)):
        for j in range(0, text_lengths[i], stride):
            # for every chunk, assign input hashes, attention mask and loc
            start, end = j, j + block_size
            done = False

            if end >= text_lengths[i]:
                end = text_lengths[i]
                start = max(end - block_size, 0)
                done = True

            if not use_subwords:
                input_hashes[current_chunk, : end - start] = flat_hashed_ids[offset + start : offset + end]
                attention_mask[current_chunk, : end - start] = 1
            else:
                chunk = [cls_token_id] + batch_of_texts[i][start:end] + [sep_token_id]
                input_ids[current_chunk, : len(chunk)] = chunk
                attention_mask[current_chunk, : len(chunk)] = 1

            locs[current_chunk, :] = [i, start, end]
            current_chunk += 1

            if done:
                break

        offset += text_lengths[i]

    assert current_chunk == num_chunks
    n_batches = math.ceil(len(attention_mask) / batch_size)

    # containers for the final logits
    all_logits = [
        np.zeros(
            (length, model.config.num_labels),
            dtype=np.float16,
        )
        for length in text_lengths
    ]
    # container for the number of chunks that any character was part of (to average chunk predictions)
    all_counts = [np.zeros(length, dtype=np.int16) for length in text_lengths]

    uses_lang_adapters = getattr(model.config, "language_adapter", "off") == "on"
    if uses_lang_adapters:
        if lang_code is None:
            raise ValueError("Please specify a `lang_code` when using a model with language adapters.")

        if isinstance(model, BertCharORTWrapper):
            raise ValueError("Language adapters are not supported in ONNX models.")

        language_ids = np.array(
            [Constants.LANG_CODE_TO_INDEX[lang_code]] * batch_size,
            dtype=int,
        )
    else:
        language_ids = None

    # forward passes through all chunks
    for batch_idx in tqdm(range(n_batches), disable=not verbose):
        start, end = batch_idx * batch_size, min(len(attention_mask), (batch_idx + 1) * batch_size)

        if not use_subwords:
            batch_input_hashes = input_hashes[start:end]
        else:
            batch_input_ids = input_ids[start:end]
        batch_attention_mask = attention_mask[start:end]

        if len(batch_attention_mask) < batch_size and pad_last_batch:
            n_missing = batch_size - len(batch_attention_mask)

            if not use_subwords:
                batch_input_hashes = np.pad(batch_input_hashes, ((0, n_missing), (0, 0), (0, 0)))
            else:
                # Pad with the specific pad_token_id for the tokenizer
                batch_input_ids = np.pad(batch_input_ids, ((0, n_missing), (0, 0)), constant_values=pad_token_id)
            batch_attention_mask = np.pad(batch_attention_mask, ((0, n_missing), (0, 0)))

        kwargs = {"language_ids": language_ids[: len(batch_attention_mask)]} if uses_lang_adapters else {}
        if use_subwords:
            kwargs["input_ids"] = batch_input_ids
        else:
            kwargs["hashed_ids"] = batch_input_hashes

        logits = model(
            attention_mask=batch_attention_mask,
            **kwargs,
        )["logits"]

        if use_subwords:
            logits = logits[:, 1:-1, :]  # remove CLS and SEP tokens

        for i in range(start, end):
            original_idx, start_char_idx, end_char_idx = locs[i]
            all_logits[original_idx][start_char_idx:end_char_idx] += logits[i - start, : end_char_idx - start_char_idx]
            all_counts[original_idx][start_char_idx:end_char_idx] += 1

    # so far, logits are summed, so we average them here
    all_logits = [(logits / counts[:, None]).astype(np.float16) for logits, counts in zip(all_logits, all_counts)]

    return (
        all_logits,
        offset_mapping if use_subwords else None,
        tokenizer if use_subwords else None,
        tokens if use_subwords else None,
    )



from dataclasses import dataclass
import json

import numpy as np
from sklearn import linear_model
from sklearn.metrics import f1_score
from transformers import AutoModelForTokenClassification, HfArgumentParser

import wtpsplit.models  # noqa
from wtpsplit.extract import PyTorchWrapper
from wtpsplit.utils import Constants
from wtpsplit.evaluation.adapt import process_logits


@dataclass
class Args:
    model_path: str
    lang: str
    block_size: int = 512
    stride: int = 64
    batch_size: int = 32
    n_subsample: int = None
    device: str = "cpu"


def load_iwslt(path, fix_space=True):
    char_labels = []

    label_dict = {
        "O": 0,
        "QUESTION": 1,
        "PERIOD": 2,
        "COMMA": 3,
    }

    text = ""

    for i, row in enumerate(open(path)):
        token, label = row.rstrip().split("\t")

        if not fix_space or (i > 0 and token not in {"'m", "n't", "'ll", "'s"}):
            text += " "
            char_labels.append(0)

        text += token
        char_labels.extend([0] * len(token))
        char_labels[-1] = label_dict[label]

    return text, np.array(char_labels)


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()

    if args.lang == "en":
        train_text, train_char_labels = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "en" / "train2012",
        )
        test_text, test_char_labels = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "en" / "test2011",
        )
    else:
        train_text, train_char_labels = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "bn" / "train",
        )
        test_text, test_char_labels = load_iwslt(
            Constants.ROOT_DIR / "data" / "external" / "punctuation_annotation" / "bn" / "test_ref",
        )

    print("Loading model...")
    model = PyTorchWrapper(AutoModelForTokenClassification.from_pretrained(args.model_path).to(args.device))

    if args.n_subsample is None:
        args.n_subsample = len(train_text)
    print("Using", args.n_subsample, "train samples...")

    train_logits = process_logits(train_text[: args.n_subsample], model, args.lang, args)

    clf = linear_model.LogisticRegression(
        penalty=None, multi_class="multinomial", max_iter=10_000, random_state=42, verbose=1
    )
    clf.fit(train_logits, train_char_labels[: args.n_subsample])

    test_logits = process_logits(test_text, model, args.lang, args)

    test_preds = clf.predict(test_logits)

    question, period, comma = (
        f1_score(test_char_labels == 1, test_preds == 1),
        f1_score(test_char_labels == 2, test_preds == 2),
        f1_score(test_char_labels == 3, test_preds == 3),
    )
    avg = np.mean([question, period, comma])
    results = {
        "C": comma,
        "P": period,
        "Q": question,
        "AVG": avg,
    }
    print(results)

    if not (Constants.CACHE_DIR / "extrinsic").exists():
        (Constants.CACHE_DIR / "extrinsic").mkdir()
    json.dump(
        results,
        open(
            Constants.CACHE_DIR / "extrinsic" / f"iwslt_{args.model_path.replace('/','_')}_{args.lang}.json",
            "w",
        ),
    )



import subprocess

SPACY_LANG_TO_DP_MODEL = {
    # "ca": "ca_core_news_sm",
    # "zh": "zh_core_web_sm",
    # "hr": "hr_core_news_sm",
    # "da": "da_core_news_sm",
    # "nl": "nl_core_news_sm",
    # "en": "en_core_web_sm",
    # "fi": "fi_core_news_sm",
    # "fr": "fr_core_news_sm",
    # "de": "de_core_news_sm",
    # "el": "el_core_news_sm",
    # "it": "it_core_news_sm",
    # "ja": "ja_core_news_sm",
    # "ko": "ko_core_news_sm",
    # "lt": "lt_core_news_sm",
    # "mk": "mk_core_news_sm",
    # "nb": "nb_core_news_sm",
    # "pl": "pl_core_news_sm",
    # "pt": "pt_core_news_sm",
    # "ro": "ro_core_news_sm",
    # "ru": "ru_core_news_sm",
    # "es": "es_core_news_sm",
    # "sv": "sv_core_news_sm",
    # "uk": "uk_core_news_sm",
    "multi": "xx_sent_ud_sm"
}


def download_models():
    for lang, model in SPACY_LANG_TO_DP_MODEL.items():
        subprocess.run(["python3", "-m", "spacy", "download", model])


if __name__ == "__main__":
    download_models()



import copy
import logging
import os
import sys
import time
from dataclasses import dataclass
from typing import List

import h5py
import numpy as np
import spacy_alignments as tokenizations
import torch
from datasets import load_dataset
from tqdm.auto import tqdm
from transformers import AutoModelForTokenClassification, AutoTokenizer, HfArgumentParser

import adapters
import wtpsplit.models  # noqa: F401
from wtpsplit.evaluation import get_labels, train_mixture
from wtpsplit.evaluation.evaluate_sepp_nlg_subtask1 import evaluate_subtask1
from wtpsplit.evaluation.adapt import process_logits
from wtpsplit.extract import PyTorchWrapper, extract
from wtpsplit.utils import Constants, sigmoid

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def get_token_labels(a, b, a_labels):
    a2b, b2a = tokenizations.get_alignments(a, b)

    b_labels = []

    for i, label_indices in enumerate(b2a):
        aligned_subwords = []

        if label_indices:
            for j in label_indices:
                if j < len(a_labels):
                    aligned_subwords.append(a_labels[j])

        if True in aligned_subwords:
            b_labels.append(1)
        else:
            b_labels.append(0)
    if not np.sum(a_labels) == sum(b_labels):
        print(np.sum(a_labels), sum(b_labels))
    b_labels[-1] = 1  # last is always 1
    return b_labels


@dataclass
class Args:
    model_path: str
    adapter_path: str = None
    # eval data in the format:
    # {
    #    "<lang_code>": {
    #        "sentence": {
    #            "<dataset_name>": {
    #                 "meta": {
    #                     "train_data": ["train sentence 1", "train sentence 2"]
    #                 },
    #                 "data": ["test sentence 1", "test sentence 2"]
    #            }
    #        }
    #    }
    # }
    eval_data_path: str = "data/ted2020.pth"
    valid_text_path: str = None  # "data/sentence/valid.parquet"
    device: str = "cpu"
    block_size: int = 512
    stride: int = 64
    batch_size: int = 32
    include_langs: List[str] = None
    include_splits: List[str] = None
    threshold: float = 0.025
    max_n_train_sentences: int = 100
    max_n_test_sentences: int = sys.maxsize
    save_suffix: str = ""
    skip_adaptation: bool = False
    clf_from_scratch: bool = False
    skip_punct: bool = True


def process_logits_and_tokens(text, model, lang_code, args):
    # variation of process_logits used in intrinsic.py for word-based evals, returning tokens as well.
    if isinstance(text, list):
        logits = []
        tokens = []
        for short_seq in tqdm(text, desc="Evaluating...", disable=False):
            current_logits, current_offsets_mapping, tokenizer, _ = extract(
                [short_seq],
                model,
                lang_code=lang_code,
                stride=args.stride,
                max_block_size=args.block_size,
                batch_size=args.batch_size,
                pad_last_batch=True,
                verbose=False,
            )
            current_logits = current_logits[0]
            if current_offsets_mapping is not None:
                current_offsets_mapping = current_offsets_mapping[0]

            current_tokens = tokenizer.encode(short_seq, verbose=False, add_special_tokens=False)

            logits.append(current_logits)
            tokens.append(current_tokens)
    else:
        raise NotImplementedError
    return logits, tokens


def load_or_compute_logits(args, model, eval_data, valid_data=None, save_str: str = None):
    logits_path = Constants.CACHE_DIR / "ted2020" / f"{save_str}.h5"

    if not os.path.exists(Constants.CACHE_DIR / "ted2020"):
        os.makedirs(Constants.CACHE_DIR / "ted2020")

    use_langs = eval_data.keys()

    total_test_time = 0  # Initialize total test processing time

    with h5py.File(logits_path, "a") as f, torch.no_grad():
        if not args.include_splits:
            splits = ["surprise_test", "test"]
        else:
            splits = args.include_splits
        for lang_code in tqdm(use_langs, desc="Languages"):
            if args.include_langs is not None and lang_code not in args.include_langs:
                continue

            if lang_code not in f:
                lang_group = f.create_group(lang_code)
            else:
                lang_group = f[lang_code]

            # eval data
            for dataset_name, dataset in tqdm(eval_data[lang_code]["sentence"].items(), desc=lang_code):
                if dataset_name not in splits:
                    continue
                try:
                    if args.adapter_path:
                        if args.clf_from_scratch:
                            model.model.classifier = torch.nn.Linear(model.model.classifier.in_features, 1)
                        # we trained adapters on "train" split but uniformly save as "surprise_test"
                        model.model.load_adapter(
                            args.adapter_path + "/" + "surprise_test" + "/" + lang_code,
                            set_active=True,
                            with_head=True,
                            load_as="text",
                        )
                    if not os.path.exists(os.path.join(args.model_path, "pytorch_model.bin")) and not os.path.exists(
                        os.path.join(args.model_path, "model.safetensors")
                    ):
                        model_path = os.path.join(args.model_path, dataset_name, "en")
                        if not os.path.exists(model_path):
                            model_path = args.model_path
                        model = PyTorchWrapper(
                            AutoModelForTokenClassification.from_pretrained(model_path).to(args.device)
                        )
                except Exception as e:
                    print(f"Error loading adapter for {dataset_name} in {lang_code}: {e}")
                    continue
                if dataset_name not in lang_group:
                    dset_group = lang_group.create_group(dataset_name)
                else:
                    dset_group = lang_group[dataset_name]

                if "test_logits" not in dset_group:
                    test_sentences = dataset["data"][: args.max_n_test_sentences]
                    if isinstance(test_sentences[0], list):
                        # short-seq eval: list of lists
                        test_text = [
                            Constants.SEPARATORS.get(lang_code, " ").join(sentence) for sentence in test_sentences
                        ]
                    else:
                        raise NotImplementedError

                    start_time = time.time()
                    test_logits, test_tokens = process_logits_and_tokens(test_text, model, lang_code, args)
                    end_time = time.time()
                    total_test_time += end_time - start_time
                    if isinstance(test_sentences[0], list):
                        test_logit_lengths = []
                        # store start and end indices for each pair, used later to slice the logits
                        # (h5py does not like different length np arrays as list elements)
                        all_logit_lengths = np.append(0, np.cumsum([len(logits) for logits in test_logits]))
                        # append tuple of start and end indices for each pair
                        for i in range(len(test_logits)):
                            test_logit_lengths.append((all_logit_lengths[i], all_logit_lengths[i + 1] - 1))
                        test_logits = np.concatenate(test_logits)
                        test_tokens = np.concatenate(test_tokens)

                        dset_group.create_dataset("test_logit_lengths", data=test_logit_lengths)

                    dset_group.create_dataset("test_logits", data=test_logits)
                    dset_group.create_dataset("test_tokens", data=test_tokens)

                train_sentences = dataset["meta"].get("train_data")
                if train_sentences is not None and "train_logits" not in dset_group and not args.skip_adaptation:
                    train_sentences = train_sentences[: args.max_n_train_sentences]
                    if isinstance(train_sentences[0], list):
                        # short-seq eval: list of lists
                        train_text = [
                            Constants.SEPARATORS.get(lang_code, " ").join(sentence) for sentence in train_sentences
                        ]
                    train_logits = process_logits(train_text, model, lang_code, args)
                    if isinstance(train_sentences[0], list):
                        train_logits = np.concatenate(train_logits)
                        train_labels = [
                            get_labels(lang_code, short_seq, after_space=False)[:-1] for short_seq in train_sentences
                        ]

                        # flatten; append 0 eos to account for later indexing/slicing
                        train_labels = np.append(np.concatenate(train_labels), 1)
                        assert len(train_labels) == len(train_logits) + 1

                    dset_group.create_dataset("train_logits", data=train_logits)
                    dset_group.create_dataset("train_labels", data=train_labels)

    end_time = time.time()
    return h5py.File(logits_path, "r"), total_test_time / 60  # to minutes


def compute_statistics(values):
    if not values:  # Check for empty values list
        return {"mean": None, "median": None, "std": None, "min": None, "min_lang": None, "max": None, "max_lang": None}

    scores, langs = zip(*values)
    min_index = np.argmin(scores)
    max_index = np.argmax(scores)
    return {
        "mean": np.mean(scores),
        "median": np.median(scores),
        "std": np.std(scores),
        "min": scores[min_index],
        "min_lang": langs[min_index],
        "max": scores[max_index],
        "max_lang": langs[max_index],
    }


def main(args):
    save_model_path = args.model_path
    if args.adapter_path:
        save_model_path = args.adapter_path
    save_str = f"{save_model_path.replace('/','_')}_b{args.block_size}_s{args.stride}"

    eval_data = torch.load(args.eval_data_path)
    if args.valid_text_path is not None:
        valid_data = load_dataset("parquet", data_files=args.valid_text_path, split="train")
    else:
        valid_data = None

    print("Loading model...")
    # if model_path does not contain a model, take first subfolder
    if not os.path.exists(os.path.join(args.model_path, "pytorch_model.bin")) and not os.path.exists(
        os.path.join(args.model_path, "model.safetensors")
    ):
        try:
            model_path = os.path.join(args.model_path, os.listdir(args.model_path)[0], "en")
        except:  # noqa
            model_path = args.model_path
        print(model_path)
    else:
        model_path = args.model_path
    model = PyTorchWrapper(AutoModelForTokenClassification.from_pretrained(model_path).to(args.device))
    if args.adapter_path:
        model_type = model.model.config.model_type
        # adapters need xlm-roberta as model type.
        model.model.config.model_type = "xlm-roberta"
        adapters.init(model.model)
        # reset model type (used later)
        model.model.config.model_type = model_type

    save_str += f"{args.save_suffix}"
    if args.max_n_test_sentences < sys.maxsize:
        save_str += f"_n{args.max_n_test_sentences}"

    # first, logits for everything.
    f, total_test_time = load_or_compute_logits(args, model, eval_data, valid_data, save_str)

    save_str += f"_u{args.threshold}"

    tokenizer = AutoTokenizer.from_pretrained(model.config.base_model)

    clfs = {}
    # now, compute the intrinsic scores.
    for lang_code, dsets in tqdm(eval_data.items()):
        if args.include_langs is not None and lang_code not in args.include_langs:
            continue

        print(f"Predicting {lang_code}...")

        for dataset_name, dataset in dsets["sentence"].items():
            if not os.path.exists(Constants.CACHE_DIR / "ted2020" / save_str / lang_code / dataset_name):
                os.makedirs(Constants.CACHE_DIR / "ted2020" / save_str / lang_code / dataset_name)
            for supervision in ["u", "t", "punct"]:
                if not os.path.exists(
                    Constants.CACHE_DIR / "ted2020" / save_str / lang_code / dataset_name / supervision
                ):
                    os.makedirs(Constants.CACHE_DIR / "ted2020" / save_str / lang_code / dataset_name / supervision)
            if "surprise" not in dataset_name:
                continue
            sentences = dataset["data"][: args.max_n_test_sentences]
            # check if f[lang_code][dataset_name] exists
            if lang_code not in f or dataset_name not in f[lang_code]:
                continue

            if "train_logits" in f[lang_code][dataset_name] and not args.skip_adaptation and "surprise" in dataset_name:
                feature_indices = None
                clf = train_mixture(
                    [lang_code],
                    f[lang_code][dataset_name]["train_logits"][:],
                    f[lang_code][dataset_name]["train_labels"][:],
                    features=feature_indices,
                    skip_punct=args.skip_punct,
                )
                if clf[0] is not None:
                    print(clf)

                clf = list(copy.deepcopy(clf))
                # 1 clf for each lang: train data is same for both.
                clfs[lang_code] = clf

            gt_dir = Constants.ROOT_DIR.parent / "data" / "sepp_nlg_2021_data" / lang_code / dataset_name
            test_files = sorted([f for f in gt_dir.glob("*.tsv") if f.is_file()])

            if isinstance(sentences[0], list):
                for i, short_seq in enumerate(sentences):
                    start, end = f[lang_code][dataset_name]["test_logit_lengths"][i]
                    current_logits = f[lang_code][dataset_name]["test_logits"][:][start : end + 1]
                    current_tokens = tokenizer.convert_ids_to_tokens(
                        f[lang_code][dataset_name]["test_tokens"][:][start : end + 1]
                    )

                    with open(test_files[i], "r", encoding="utf-8") as file:
                        idx = test_files[i].name.split(".")[0]
                        lines = file.read().strip().split("\n")
                        rows = [line.split("\t") for line in lines]
                        gt_words = [row[0] for row in rows]
                        gt_labels = [row[1] for row in rows]
                        if gt_labels[-1] != "1":
                            print("0 label at end!")
                    u_preds = sigmoid(current_logits[:, 0]) > args.threshold
                    if not args.skip_adaptation:
                        t_preds = sigmoid(current_logits[:, 0]) > clfs[lang_code][-1]
                    else:
                        t_preds = None
                    if not args.skip_adaptation and not args.skip_punct:
                        punct_preds = clfs[lang_code][0].predict_proba(current_logits)[:, 1] > clf[2]
                    else:
                        punct_preds = None

                    # write to tsv as per the challenge reqs
                    # can then be evaluated via evaluate_sepp_nlg_subtask1.py
                    for supervision, preds in zip(["u", "t", "punct"], [u_preds, t_preds, punct_preds]):
                        if preds is None:
                            continue

                        word_labels = get_token_labels(current_tokens, gt_words, preds)

                        with open(
                            Constants.CACHE_DIR
                            / "ted2020"
                            / save_str
                            / lang_code
                            / dataset_name
                            / supervision
                            / f"{idx}.tsv",
                            "w",
                            encoding="utf-8",
                        ) as file:
                            for word, label in zip(gt_words, word_labels):
                                file.write(f"{word}\t{label}\n")
            else:
                raise NotImplementedError
    return total_test_time, save_str


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()
    total_test_time, save_str = main(args)
    print(total_test_time)
    supervision = ["u", "t", "punct"]
    if args.skip_adaptation:
        supervision.remove("t")
    if args.skip_punct:
        supervision.remove("punct")
    if not args.include_langs:
        include_langs = ["en", "de", "fr", "it"]
    else:
        include_langs = args.include_langs
    if not args.include_splits:
        include_splits = ["surprise_test", "test"]
    else:
        include_splits = args.include_splits
    results = evaluate_subtask1(include_splits, include_langs, save_str, supervision, args.max_n_test_sentences)



import logging
import os
import torch
import torch.nn as nn
from wtpsplit.utils import Constants

logger = logging.getLogger(__name__)


class Model(nn.Module):
    def __init__(
        self,
        backbone,
        loss_margin=0.5,
        use_loss_weights=False,
        do_sentence_training=True,
        do_auxiliary_training=False,
        aux_training_weight=1.0,
    ):
        super().__init__()
        self.backbone = backbone
        self.config = self.backbone.config

        assert loss_margin <= 0.5

        self.loss_margin = loss_margin
        self.use_loss_weights = use_loss_weights
        self.do_sentence_training = do_sentence_training
        self.do_auxiliary_training = do_auxiliary_training
        self.aux_training_weight = aux_training_weight

    @property
    def device(self):
        return self.backbone.device

    def forward(
        self,
        input_ids,
        language_ids=None,
        attention_mask=None,
        position_ids=None,
        labels=None,
        label_weights=None,
        **kwargs,
    ):
        if position_ids is not None:
            # XXX: 1 is pad token id
            if "xlm" in self.config.model_type:
                reduced_attention_mask = (input_ids != 1).to(torch.long)
            else:
                reduced_attention_mask = (input_ids != 0).to(torch.long)

        output = dict(
            self.backbone.forward(
                input_ids=input_ids,
                language_ids=language_ids,
                attention_mask=attention_mask,
                position_ids=position_ids,
                **kwargs,
            )
        )
        logits = output["logits"]

        if labels is not None:
            loss_fn = nn.BCEWithLogitsLoss(reduction="none")

            # main (newline prediction) objective
            if self.do_sentence_training:
                # label smoothing
                sentence_labels = (0.5 - self.loss_margin) + (labels == Constants.NEWLINE_INDEX + 1).to(
                    logits.dtype
                ).view(-1) * self.loss_margin * 2
                sentence_logits = logits[:, :, Constants.NEWLINE_INDEX].view(-1)

                loss = (
                    loss_fn(
                        sentence_logits,
                        sentence_labels,
                    )
                    * (label_weights.view(-1) if label_weights is not None and self.use_loss_weights else 1)
                    * reduced_attention_mask.view(-1)
                ).sum() / reduced_attention_mask.sum()

            # auxiliary (punctuation prediction) objective
            if self.do_auxiliary_training:
                loss_fn = nn.CrossEntropyLoss()

                # exclude newline and no labels
                aux_labels = torch.where(
                    (labels == 0) | (labels == Constants.NEWLINE_INDEX + 1),
                    0,
                    labels - Constants.AUX_OFFSET,
                )
                # exclude reduced_attention_mask tokens from labels
                aux_labels = torch.where(
                    reduced_attention_mask == 1,
                    aux_labels,
                    loss_fn.ignore_index,
                )

                aux_loss = loss_fn(
                    logits[:, :, Constants.AUX_OFFSET :].view(-1, self.config.num_labels - Constants.AUX_OFFSET),
                    aux_labels.view(-1),
                )

                loss = loss + self.aux_training_weight * aux_loss

            output["loss"] = loss

        return output


def cleanup_cache_files(datasets) -> int:
    """Clean up all cache files in the dataset cache directory, except those currently used by any of the provided datasets.

    Args:
        datasets (List[Dataset]): A list of dataset objects.

    Be careful when running this command that no other process is currently using other cache files.

    Returns:
        int: Number of removed files.
    """
    if not datasets:
        return 0

    # Collect all current cache files from the provided datasets
    current_cache_files = set()
    for dataset in datasets:
        dataset_cache_files = [os.path.abspath(cache_file["filename"]) for cache_file in dataset.cache_files]
        current_cache_files.update(dataset_cache_files)
    logger.warning(f"Found {len(current_cache_files)} cache files used by the provided datasets.")

    if not current_cache_files:
        return 0

    # Assuming all datasets have cache files in the same directory
    cache_directory = os.path.dirname(next(iter(current_cache_files)))

    files = os.listdir(cache_directory)
    files_to_remove = []
    for f_name in files:
        full_name = os.path.abspath(os.path.join(cache_directory, f_name))
        if f_name.startswith("cache-") and f_name.endswith(".arrow") and full_name not in current_cache_files:
            files_to_remove.append(full_name)

    for file_path in files_to_remove:
        logger.warning(f"Removing {file_path}")
        os.remove(file_path)

    return len(files_to_remove)



import math
import random
import sys
from collections import defaultdict
from dataclasses import dataclass
from itertools import cycle
from typing import Iterable, List, Sequence, Tuple

import numpy as np
import torch
import transformers
from datasets import Dataset
from torch.utils.data import BatchSampler, ConcatDataset, DataLoader, SubsetRandomSampler
from tqdm import tqdm
from transformers import AutoTokenizer, HfArgumentParser, Trainer, TrainerCallback, TrainingArguments

import wandb
from wtpsplit.models import SubwordXLMForTokenClassification
from wtpsplit.utils import Constants


@dataclass
class Args:
    block_size: int = 256
    num_layers: int = 12  # number of layers
    lim_lookahead: bool = False  # our "Lookahead" ablation
    without_pretraining: bool = False  # our "No pre-training" ablation
    no_sm_corruption: bool = False  # our "Only clean text" ablation


# Parsing command line arguments or JSON config files as needed
parser = HfArgumentParser([Args, TrainingArguments])

if len(sys.argv) > 1 and sys.argv[1].endswith(".json"):
    args, training_args = parser.parse_json_file(sys.argv[1])
else:
    args, training_args = parser.parse_args_into_dataclasses()

data_path = "data/all_data_11_05-all.pth"
all_data = torch.load(data_path)

block_size = args.block_size

train_sentences = defaultdict(lambda: defaultdict(list))
test_sentences = defaultdict(lambda: defaultdict(list))

punct_chars = set(Constants.PUNCTUATION_CHARS)


for lang_code in tqdm(all_data, desc="Loading data"):
    if "-" in lang_code or "_" in lang_code:
        # we only train on monolingual data in SM, so no "en-de" code-switching for example!
        pass
    elif (
        "ud" in all_data[lang_code]["sentence"]
        and all_data[lang_code]["sentence"]["ud"]["meta"]["train_data"] is not None
    ):
        train_data = all_data[lang_code]["sentence"]["ud"]["meta"]["train_data"]

        if len(train_data) < 10000:
            # some languages have an insufficient number of sentences to fill a single batch
            # this is just a quick way to upsample these so we don't run into problems later
            # later we will use a uniform round-robin sampler for all languages
            train_data = train_data * (10000 // len(train_data) + 1)

        train_sentences[lang_code]["uncorrupted"].extend(train_data)

        if not args.no_sm_corruption:
            train_data = all_data[lang_code]["sentence"]["ud-corrupted-asr"]["meta"]["train_data"]

            if len(train_data) < 5000:
                # some languages have an insufficient number of sentences to fill a single batch
                # this is just a quick way to upsample these so we don't run into problems later
                # later we will use a uniform round-robin sampler for all languages
                train_data = train_data * (10000 // len(train_data) + 1)

            train_sentences[lang_code]["corrupted-asr"].extend(train_data)

            train_data = all_data[lang_code]["sentence"]["ud-corrupted-social-media"]["meta"]["train_data"]

            if len(train_data) < 5000:
                # some languages have an insufficient number of sentences to fill a single batch
                # this is just a quick way to upsample these so we don't run into problems later
                # later we will use a uniform round-robin sampler for all languages
                train_data = train_data * (10000 // len(train_data) + 1)

            train_sentences[lang_code]["corrupted-social-media"].extend(train_data)

    elif (
        "opus100" in all_data[lang_code]["sentence"]
        and all_data[lang_code]["sentence"]["opus100"]["meta"]["train_data"] is not None
    ):
        train_data = all_data[lang_code]["sentence"]["opus100"]["meta"]["train_data"]
        train_sentences[lang_code]["uncorrupted"].extend(train_data)

        if not args.no_sm_corruption:
            train_data = all_data[lang_code]["sentence"]["opus100-corrupted-asr"]["meta"]["train_data"]
            train_sentences[lang_code]["corrupted-asr"].extend(train_data)

            train_data = all_data[lang_code]["sentence"]["opus100-corrupted-social-media"]["meta"]["train_data"]
            train_sentences[lang_code]["corrupted-social-media"].extend(train_data)
    else:
        train_data = all_data[lang_code]["sentence"]["nllb"]["meta"]["train_data"]
        train_sentences[lang_code]["uncorrupted"].extend(train_data)

        if not args.no_sm_corruption:
            train_data = all_data[lang_code]["sentence"]["nllb-corrupted-asr"]["meta"]["train_data"]
            train_sentences[lang_code]["corrupted-asr"].extend(train_data)

            train_data = all_data[lang_code]["sentence"]["nllb-corrupted-social-media"]["meta"]["train_data"]
            train_sentences[lang_code]["corrupted-social-media"].extend(train_data)

    for dataset in all_data[lang_code]["sentence"]:
        if any(dataset.startswith(x) for x in ["short-sequences", "legal"]):
            continue

        test_data = all_data[lang_code]["sentence"][dataset]["data"]
        test_sentences[lang_code][dataset].extend(test_data[:200])


tokenizer_checkpoint = "xlm-roberta-base"

if args.without_pretraining:
    model_checkpoint = "xlm-roberta-base"
elif args.num_layers == 1:
    if not args.lim_lookahead:
        model_checkpoint = "segment-any-text/sat-1l-no-limited-lookahead"
    else:
        model_checkpoint = "segment-any-text/sat-1l"
elif args.num_layers == 3:
    if not args.lim_lookahead:
        model_checkpoint = "segment-any-text/sat-3l-no-limited-lookahead"
    else:
        model_checkpoint = "segment-any-text/sat-3"
elif args.num_layers == 6:
    if not args.lim_lookahead:
        model_checkpoint = "segment-any-text/sat-6l-no-limited-lookahead"
    else:
        model_checkpoint = "segment-any-text/sat-6l"
elif args.num_layers == 9:
    if not args.lim_lookahead:
        model_checkpoint = "segment-any-text/sat-9l-no-limited-lookahead"
    else:
        model_checkpoint = "segment-any-text/sat-9l"
elif args.num_layers == 12:
    if not args.lim_lookahead:
        model_checkpoint = "segment-any-text/sat-12l-no-limited-lookahead"
    else:
        model_checkpoint = "segment-any-text/sat-12l"
else:
    raise ValueError("Invalid number of layers. Valid values are 1, 3, 6, 9, 12.")

print(model_checkpoint)

tokenizer = AutoTokenizer.from_pretrained(tokenizer_checkpoint)
assert isinstance(tokenizer, transformers.PreTrainedTokenizerFast)

if args.num_layers == 3 and args.without_pretraining:
    # special case for one of our ablations, where we trim XLM-R (without any of our newline pretraining) to 3 layers
    model = SubwordXLMForTokenClassification.from_pretrained(
        model_checkpoint,
        num_labels=1,
        ignore_mismatched_sizes=True,
        num_hidden_layers=3,
    )
else:
    model = SubwordXLMForTokenClassification.from_pretrained(
        model_checkpoint,
        num_labels=1,
        ignore_mismatched_sizes=True,
    )


def tokenize_and_get_labels(sentences, lang_code, dataset_name):
    separator = Constants.SEPARATORS.get(lang_code, " ")

    joined_sentence = ""
    sentence_start_positions = []
    current_position = 0

    for sentence in sentences:
        if random.random() < 0.1 and sentence[-1] in punct_chars and dataset_name == "corrupted-social-media":
            if separator == " ":
                separator_used = ""
            else:
                separator_used = " "
        else:
            separator_used = separator

        if joined_sentence:
            joined_sentence += separator_used
            current_position += len(separator_used)
        start_position = current_position
        joined_sentence += sentence
        current_position += len(sentence)
        sentence_start_positions.append(start_position + len(sentence) - 1)

    tokenized_input = tokenizer(
        joined_sentence,
        return_offsets_mapping=True,
        add_special_tokens=False,
        truncation=False,
    )

    tokens = tokenized_input.tokens()
    offsets = tokenized_input["offset_mapping"]
    sentence_ending_labels = [0] * len(tokens)

    sentence_ending_labels[-1] = 1
    sentence_index = 0

    for i in range(len(offsets)):
        if offsets[i][0] > sentence_start_positions[sentence_index]:
            sentence_ending_labels[i - 1] = 1
            sentence_index += 1

    input_ids = [0] + tokenized_input["input_ids"] + [2]
    labels = [0] + sentence_ending_labels + [0]

    return input_ids, labels


def pack_sentences(input_data_dict, block_size):
    packed_data = defaultdict(lambda: defaultdict(lambda: {"input_ids": [], "attention_mask": [], "labels": []}))

    for lang_code in tqdm(input_data_dict):
        for dataset_name, sentences in input_data_dict[lang_code].items():
            if dataset_name == "corrupted-social-media":
                p_add_to_block = 0.5
            else:
                p_add_to_block = 1.0

            token_count, one_block_sentences = 0, []

            for sentence in sentences:
                num_sentence_tokens = len(tokenizer(sentence, add_special_tokens=False)["input_ids"])

                if not sentence or sentence.isnumeric() or num_sentence_tokens == 0:
                    continue

                if token_count + num_sentence_tokens < block_size - 4 and (
                    random.random() <= p_add_to_block or len(one_block_sentences) == 0
                ):
                    one_block_sentences.append(sentence)
                    token_count += num_sentence_tokens
                else:
                    if one_block_sentences:
                        input_ids, labels = tokenize_and_get_labels(one_block_sentences, lang_code, dataset_name)

                        num_to_pad = block_size - len(input_ids)
                        attention_mask = [1] * len(input_ids) + [0] * num_to_pad
                        input_ids += [tokenizer.pad_token_id] * num_to_pad
                        labels += [-100] * num_to_pad

                        assert len(input_ids) == block_size, len(input_ids)
                        assert len(input_ids) == len(labels), (
                            len(input_ids),
                            len(labels),
                        )

                        packed_data[lang_code][dataset_name]["input_ids"].append(input_ids)
                        packed_data[lang_code][dataset_name]["attention_mask"].append(attention_mask)
                        packed_data[lang_code][dataset_name]["labels"].append(labels)

                    if num_sentence_tokens > block_size - 4:
                        one_block_sentences = []
                        token_count = 0
                    else:
                        one_block_sentences = [sentence]
                        token_count = num_sentence_tokens

            if one_block_sentences:
                input_ids, labels = tokenize_and_get_labels(one_block_sentences, lang_code, dataset_name)

                num_to_pad = block_size - len(input_ids)
                attention_mask = [1] * len(input_ids) + [0] * num_to_pad
                input_ids += [tokenizer.pad_token_id] * num_to_pad
                labels += [-100] * num_to_pad

                assert len(input_ids) == block_size, len(input_ids)
                assert len(input_ids) == len(labels), (len(input_ids), len(labels))

                packed_data[lang_code][dataset_name]["input_ids"].append(input_ids)
                packed_data[lang_code][dataset_name]["attention_mask"].append(attention_mask)
                packed_data[lang_code][dataset_name]["labels"].append(labels)

            assert len(packed_data[lang_code][dataset_name]["input_ids"]) == len(
                packed_data[lang_code][dataset_name]["labels"]
            )

    return packed_data


packed_train_data = pack_sentences(train_sentences, block_size)
packed_test_data = pack_sentences(test_sentences, block_size)
test_dataset = {lang_code: defaultdict(dict) for lang_code in packed_test_data}

for lang_code in packed_test_data:
    for dataset_name in packed_test_data[lang_code]:
        test_dataset[lang_code][dataset_name] = Dataset.from_dict(packed_test_data[lang_code][dataset_name])

experiment_name = model_checkpoint.split("/")[-1]

if args.no_sm_corruption:
    experiment_name += "-no-corruption"

training_args.output_dir = experiment_name


def compute_prf(true_values, predicted_values):
    TP = np.sum((predicted_values == 1) & (true_values == 1))
    FP = np.sum((predicted_values == 1) & (true_values == 0))
    FN = np.sum((predicted_values == 0) & (true_values == 1))

    precision = TP / (TP + FP) if (TP + FP) > 0 else 0
    recall = TP / (TP + FN) if (TP + FN) > 0 else 0
    f1_score = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

    return precision, recall, f1_score


def sigmoid_array(x):
    return 1 / (1 + np.exp(-x))


def compute_metrics(p):
    predictions, labels = p

    predictions = np.reshape(predictions, (-1,))
    labels = np.reshape(labels, (-1,))

    predictions = sigmoid_array(predictions)

    predictions = predictions[labels != -100]
    labels = labels[labels != -100]

    threshold = 0.25

    preds = (predictions > threshold).astype(int)

    precision, recall, f1 = compute_prf(labels, preds)

    output_dict = {
        "precision": precision,
        "recall": recall,
        "f1": f1,
    }

    return output_dict


class MultiDatasetEvalCallback(TrainerCallback):
    def __init__(self, eval_datasets):
        self.eval_datasets = eval_datasets

    def on_evaluate(self, args, state, control, metrics=None, **kwargs):
        pass

    def on_step_end(self, args, state, control, **kwargs):
        if state.global_step % args.eval_steps == 0:
            for lang_code in self.eval_datasets:
                for dataset_name, eval_dataset in self.eval_datasets[lang_code].items():
                    metrics = trainer.evaluate(eval_dataset)
                    for metric, result in metrics.items():
                        wandb.log(
                            {
                                f"eval/{dataset_name}/{lang_code}/{metric}": result,
                                "train/global_step": state.global_step,
                            }
                        )


multi_dataset_eval_callback = MultiDatasetEvalCallback(test_dataset)

train_datasets = []

for lang_code in packed_train_data:
    for dataset_name in packed_train_data[lang_code]:
        train_datasets.append(Dataset.from_dict(packed_train_data[lang_code][dataset_name]))

random.shuffle(train_datasets)

train_datasets = ConcatDataset(train_datasets)

run = wandb.init(project="sentence")
wandb.run.name = experiment_name

# args = TrainingArguments(
#     output_dir=experiment_name,
#     overwrite_output_dir=True,
#     evaluation_strategy="steps",
#     eval_steps=250,
#     report_to="wandb",
#     learning_rate=3e-5,
#     warmup_steps=500,
#     per_device_train_batch_size=128,
#     per_device_eval_batch_size=128,
#     weight_decay=0.01,
#     push_to_hub=False,
#     save_total_limit=1,
#     save_strategy="steps",
#     save_steps=1000,
#     load_best_model_at_end=False,
#     max_steps=20000,
# )


class RoundRobinSampler:
    def __init__(self, samplers: Sequence[Iterable], reinit: bool = False):
        self.samplers = samplers
        self.reinit = reinit

    def __iter__(self):
        iterators = [iter(sampler) for sampler in self.samplers]

        for i in cycle(range(len(iterators))):
            it = iterators[i]

            try:
                yield next(it)

            except StopIteration:
                if not self.reinit:
                    break

                it = iter(self.samplers[i])
                iterators[i] = it
                yield next(it)


def get_subset(length: int, i: int, k: int, offset: int = 0) -> Tuple[int, int]:
    assert i < k
    s = math.ceil(length / k)  # size of one split
    start = i * s
    end = min((i + 1) * s, length)
    return offset + start, offset + end


class DistributedRoundRobinBatchSampler:
    def __init__(
        self,
        lengths: List[int],
        batch_size: int,
        rank: int,
        num_replicas: int,
        drop_last: bool = False,
        seed: int = 0,
        shuffle: bool = True,
        reinit: bool = False,
    ):
        self.lengths = lengths
        offsets = [sum(lengths[:i]) for i in range(len(lengths))]
        self.ranges = [get_subset(length, rank, num_replicas, offset) for offset, length in zip(offsets, lengths)]
        self.seed = seed
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.epoch = 0
        self.reinit = reinit
        self.batch_size = batch_size
        self.batch_start = 0

    def __iter__(self):
        g = torch.Generator()
        g.manual_seed(self.seed + self.epoch)

        batch_samplers = [
            BatchSampler(
                (SubsetRandomSampler(range(start, end), generator=g) if self.shuffle else range(start, end)),
                self.batch_size,
                self.drop_last,
            )
            for (start, end) in self.ranges
        ]

        sampler = RoundRobinSampler(batch_samplers, reinit=self.reinit)
        return iter(sampler)

    def __len__(self):
        return min(length for length in self.lengths) // self.batch_size


class CustomTrainer(Trainer):
    def get_train_dataloader(self) -> DataLoader:
        dataset = self.train_dataset

        if isinstance(dataset, ConcatDataset):
            sizes = [len(ds) for ds in dataset.datasets]
        else:
            sizes = [len(dataset)]

        loader = DataLoader(
            dataset,
            batch_sampler=DistributedRoundRobinBatchSampler(
                lengths=sizes,
                batch_size=self.args.train_batch_size,
                drop_last=False,
                rank=self.args.process_index,
                num_replicas=self.args.world_size,
                seed=self.args.seed,
                reinit=True,
            ),
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
            collate_fn=self.data_collator,
        )
        return loader


trainer = CustomTrainer(
    model=model,
    args=training_args,
    train_dataset=train_datasets,
    eval_dataset=None,
    compute_metrics=compute_metrics,
    tokenizer=tokenizer,
    callbacks=[multi_dataset_eval_callback],
)

trainer.train()
